import os
import io
import csv
import json
import logging
from deriva.core import json_item_handler
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError
from deriva.transfer.download.processors.transform.base_transform_processor import BaseTransformProcessor

logger = logging.getLogger(__name__)
csv.register_dialect('unix_comma_quotemin', delimiter=',', lineterminator='\n', quoting=csv.QUOTE_MINIMAL)


class JSONtoCSVTransformProcessor(BaseTransformProcessor):
    """ Convert JSON or JSON-Stream to CSV transform processor.
    """
    def __init__(self, envars=None, **kwargs):
        super(JSONtoCSVTransformProcessor, self).__init__(envars, **kwargs)
        self._create_input_output_paths()
        self.writer = None

        # get custom params
        self.include_header = self.parameters.get('include_header', True)
        self.csv_dialect = self.parameters.get('csv_dialect', "unix_comma_quotemin")

    def process(self):
        """Reads a JSON or JSON-Stream input file, converts to CSV output with optional header row.
        """
        try:
            with io.open(self.output_abspath, mode='w', encoding='utf-8', newline='') as output_file:
                def row_handler(item):
                    if not self.writer:
                        self.writer = csv.DictWriter(output_file, fieldnames=item.keys(), dialect=self.csv_dialect)
                        if self.include_header:
                            self.writer.writeheader()
                    self.writer.writerow(item)

                json_item_handler(self.input_abspath, row_handler)

        except IOError as e:
            raise DerivaDownloadError("JSONtoCSV transform failed", e)

        return super(JSONtoCSVTransformProcessor, self).process()







import sys
from importlib import import_module
from deriva.core import read_config, format_exception
from deriva.transfer.upload import DerivaUploadConfigurationError
from deriva.transfer.upload.processors.base_processor import BaseProcessor
from deriva.transfer.upload.processors.archive_processor import BagArchiveProcessor
from deriva.transfer.upload.processors.rename_processor import FileRenameProcessor
from deriva.transfer.upload.processors.logging_processor import LoggingProcessor
from deriva.transfer.upload.processors.metadata_update_processor import MetadataUpdateProcessor

DEFAULT_PROCESSORS = {
    "BagArchiveProcessor": BagArchiveProcessor,
    "FileRenameProcessor": FileRenameProcessor,
    "LoggingProcessor": LoggingProcessor,
    "MetadataProcessor": MetadataUpdateProcessor
}


def find_processor(processor_name, processor_type=None, defaults=DEFAULT_PROCESSORS.copy(), **kwargs):
    if not processor_type:
        if processor_name in defaults:
            return defaults[processor_name]
        else:
            raise DerivaUploadConfigurationError("Unsupported processor type: %s" % processor_name)

    if not _is_processor_whitelisted(processor_type, **kwargs):
        raise DerivaUploadConfigurationError(
            "Unknown external processor type [%s]: this processor must be added to the whitelist." % processor_type)

    clazz = None
    try:
        module_name, class_name = processor_type.rsplit(".", 1)
        try:
            module = sys.modules[module_name]
        except KeyError:
            module = import_module(module_name)
        clazz = getattr(module, class_name) if module else None
    except (ImportError, AttributeError):
        pass
    if not clazz:
        raise DerivaUploadConfigurationError("Unable to import specified processor class: %s" % processor_type)

    if not issubclass(clazz, BaseProcessor):
        raise DerivaUploadConfigurationError(format_exception(
            NotImplementedError("The imported class %s is not a subclass of %s" %
                                (processor_type, BaseProcessor.__module__ + "." + BaseProcessor.__name__))))

    return clazz


def _is_processor_whitelisted(processor_type, **kwargs):
    if kwargs.get("bypass_whitelist", False):
        return True
    config_file_path = kwargs.get("config_file_path")
    config = read_config(config_file=config_file_path)
    whitelist = config.get("upload_processor_whitelist", [])

    return processor_type in whitelist



import re
import sys
from packaging import version


def get_installed_version(ver):
    """
    Generates an external version string from an internal one. This is currently only being used to determine
    whether we are operating in a "frozen" environment or not, but other decorators could be added.
    """
    version_str = str(ver)
    return version_str if not getattr(sys, 'frozen', False) else version_str + '-frozen'


def is_compatible(source_version, compat_versions):
    """
    Compare a source version string to a set of target version string specifications. Supports semantic versioning
    comparison via setuptools version comparison logic (http://setuptools.readthedocs.io/en/latest/setuptools.html#id7).

    :param source_version: a source version string
    :param compat_versions:
        an array of tuples with each tuple consisting of a set of strings of the form
        `<operator><version>`. The source_version is evaluated in a conjunction against each
        `<operator><version>` string in the tuple, using the packaging module version comparison
        function. The result of every tuple evaluation is then evaluated in a disjunction
        against other tuples in the array.  If any one of the tuples evaluates to True, then the
        returned disjunction is logically true, and the source version is assumed to be compatible.
    :return: boolean indicating compatibility

    Example 1:
    ::
        is_compatible("0.1.0", [[">=0.1.0", "<1.0.0"]])

    :return: `True`

    Example 2:
    ::
        is_compatible("1.1.0", [[">=0.1.0", "<1.0.0"]])

    :return: `False`

    Example 3:
    ::
        is_compatible("0.5.1-dev", [[">=0.3.5rc4", "<0.5.9-beta", "!=0.5.0"]])

    :return: `True`

    Example 4:
    ::
        is_compatible("0.7.7", [[">=0.7.0", "<0.7.6"], ["!=0.7.1"]])

    :return: `True` (disjunct of "!=0.7.1" cancelled out range ">=0.7.0", "<0.7.6" -- see below)

    Example 5:
    ::
        is_compatible("0.7.7", [[">=0.7.0", "<0.7.6", "!=0.7.1"]])

    :return: `False`

    Example 5:
    ::
        is_compatible("version-3.3", [["==version-3.0"], ["==version-3.1"], ["==version-3.3"]])

    :return: `True`

    """
    pattern = "^.*(?P<operator>(>=|<=|>|<|==|!=))(?P<version>.*)$"
    compat = None
    for version_spec in compat_versions:
        check = None
        for ver in version_spec:
            match = re.search(pattern, ver)
            if match:
                gd = match.groupdict()
                operator = gd["operator"]
                target_version = gd["version"]
                ret = eval("version.parse(source_version) %s version.parse(target_version)" % operator)
                if check is None:
                    check = ret
                else:
                    check = ret and check
        if compat is None:
            compat = check
        else:
            compat = check or compat

    return compat if compat is not None else False



import sys
import json
import re
from deriva.core import ErmrestCatalog, AttrDict, get_credential
from deriva.config.base_config import BaseSpec, BaseSpecList, ConfigUtil, ConfigBaseCLI
from deriva.config.annotation_config import AttrSpec, AttrSpecList

MY_VERSION = 0.99


class Annotations:
    def __init__(self, server, catalog, credentials, config):
        self.annotations = {}
        self.ignored_schema_patterns = []
        ip = config.get("ignored_schema_patterns")
        if ip is not None:
            for p in ip:
                self.ignored_schema_patterns.append(re.compile(p))
        self.types = set()
        self.managed_attributes = []
        self.ignore_unmanaged = True
        self.ignored_attributes = []
        self.consolidated_annotations = {}
        self.annotations_to_delete = []
        if config is not None:
            known_attributes = config.get("known_attributes")
            if known_attributes is not None:
                self.managed_attributes = known_attributes.get("managed", [])
                self.ignored_attributes = known_attributes.get("ignored", [])
                self.annotations_to_delete = known_attributes.get("to_delete", [])
                self.ignore_unmanaged = known_attributes.get("ignore_all_unmanaged", True)
                self.annotations["known_attributes"] = known_attributes
            else:
                self.annotations["known_attributes"] = {'managed': [], 'ignore_all_unmanaged': True}
            self.consolidated_annotations = config.get("consolidated_annotations", {})
            for k in AttrSpecList.SPEC_TYPES:
                d = self.consolidated_annotations.get(k)
                if d is None:
                    d = [dict()]
                self.consolidated_annotations[k] = AttrSpecList(known_attributes, d)
                self.annotations[k] = self.munge_specs(self.consolidated_annotations[k])

        for k in self.managed_attributes:
            if k in self.annotations_to_delete:
                raise ValueError("{k} is both 'managed' and 'to_delete'".format(k=k))
        self.catalog = ErmrestCatalog('https', server, catalog, credentials)
        self.catalog_config = self.catalog.getCatalogModel()
        if self.catalog_config.annotations is not None:
            self.add_catalog_annotations(self.catalog_config)
        if self.catalog_config.schemas is not None:
            for s in self.catalog_config.schemas.values():
                self.add_schema_annotations(s)

    def munge_specs(self, annotation_list):
        speclist = []
        if annotation_list is not None:
            if isinstance(annotation_list, AttrSpecList):
                annotation_list = annotation_list.get_specs()
            for spec in annotation_list:
                speclist.append(spec.config_format())
        return speclist

    def consolidated_schema_annotation(self, annotation):
        matches = []
        for c in self.consolidated_annotations["schema_annotations"].get_specs():
            if c.schema_entry_matches(annotation.get("schema"), key=annotation.get("uri")):
                matches.append(c)
        return self.check_consolidation(matches, annotation.get("value"))

    def consolidated_table_annotation(self, annotation):
        matches = []
        for c in self.consolidated_annotations["table_annotations"].get_specs():
            if c.table_entry_matches(annotation.get("schema"),
                                     annotation.get("table"),
                                     key=annotation.get("uri")):
                matches.append(c)
        return self.check_consolidation(matches, annotation.get("value"))

    def consolidated_column_annotation(self, annotation):
        matches = []
        for c in self.consolidated_annotations["column_annotations"].get_specs():
            if c.column_entry_matches(annotation.get("schema"),
                                      annotation.get("table"),
                                      annotation.get("column"),
                                      key=annotation.get("uri")):
                matches.append(c)
        return self.check_consolidation(matches, annotation.get("value"))

    def consolidated_foreign_key_annotation(self, annotation):
        matches = []
        for c in self.consolidated_annotations["foreign_key_annotations"].get_specs():
            if c.foreign_key_entry_matches(annotation.get("schema"),
                                           annotation.get("table"),
                                           annotation.get("foreign_key_schema"),
                                           annotation.get("foreign_key"),
                                           key=annotation.get("uri")):
                matches.append(c)
        return self.check_consolidation(matches, annotation.get("value"))

    def check_consolidation(self, matches, value):
        if len(matches) != 1:
            # Zero or more than one matching pattern, so we need the exact spec to disambiguate
            return False

        match = matches[0]

        #        if match.get("override") == True:
        #            # We don't care what the original version was. We want to go with the pattern match
        #            return True

        return match.get("value") == value

    def add_catalog_annotations(self, catalog):
        annotations = self.find_relevant_annotations(catalog.annotations)
        if annotations is not None:
            for v in annotations:
                self.annotations["catalog_annotations"].append(v)

    def add_schema_annotations(self, schema):
        annotations = self.find_relevant_annotations(schema.annotations)
        if annotations is not None:
            for v in annotations:
                v["schema"] = schema.name
                if not self.consolidated_schema_annotation(v):
                    self.annotations["schema_annotations"].append(v)
        for table in schema.tables.values():
            self.add_table_annotations(table)

    def add_table_annotations(self, table):
        annotations = self.find_relevant_annotations(table.annotations)
        if annotations is not None:
            for v in annotations:
                v["schema"] = table.schema.name
                v["table"] = table.name
                if not self.consolidated_table_annotation(v):
                    self.annotations["table_annotations"].append(v)
        for column in table.column_definitions:
            self.add_column_annotations(table, column)
        for fkey in table.foreign_keys:
            self.add_foreign_key_annotations(fkey)

    def add_column_annotations(self, table, column):
        annotations = self.find_relevant_annotations(column.annotations)
        if annotations is not None:
            for v in annotations:
                v["schema"] = table.schema.name
                v["table"] = table.name
                v["column"] = column.name
                if not self.consolidated_column_annotation(v):
                    self.annotations["column_annotations"].append(v)

    def add_foreign_key_annotations(self, fkey):
        annotations = self.find_relevant_annotations(fkey.annotations)
        if annotations is not None:
            if len(fkey.names) < 1:
                raise ValueError("foreign key without a name")
            for v in annotations:
                v["schema"] = fkey.table.schema.name
                v["table"] = fkey.table.name
                v["foreign_key_schema"] = fkey.names[0][0]
                v["foreign_key"] = fkey.names[0][1]
                if not self.consolidated_foreign_key_annotation(v):
                    self.annotations["foreign_key_annotations"].append(v)

    def find_relevant_annotations(self, annotations):
        if annotations is None or len(annotations) == 0:
            return None
        new = []
        if self.managed_attributes is None:
            for k in annotations.keys():
                if k not in self.annotations_to_delete:
                    new.append({"uri": k, "value": annotations[k]})
                    self.types.add(k)
        else:
            for k in annotations.keys():
                if k in self.managed_attributes:
                    new.append({"uri": k, "value": annotations[k]})
                    self.types.add(k)
        if len(new) == 0:
            return None
        return new

    def dumps(self):
        return json.dumps(self.annotations, indent=4, sort_keys=True)

    def types_list(self):
        types = list(self.types)
        types.sort()
        return types


def main():
    cli = ConfigBaseCLI("annotation dump tool", None, version=MY_VERSION)
    cli.parser.add_argument('-l', help="list tags encountered", action="store_true")
    args = cli.parse_cli()
    managed_attrs = None
    if args.config_file is None:
        print("No config file specified")
        return 1
    if args.host is None:
        print("No host specified")
        return 1
    config = json.load(open(args.config_file))
    credentials = get_credential(args.host, args.credential_file)
    annotations = Annotations(args.host, args.catalog, credentials, config)
    if args.l:
        for t in annotations.types_list():
            print(t)
    if not args.l:
        print(annotations.dumps())
    return 0


if __name__ == '__main__':
    sys.exit(main())



import io
import os
import sys
import shutil
import errno
import json
import math
import platform
import logging
import requests
import portalocker
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from collections import OrderedDict
from distutils import util as du_util
from urllib.parse import quote as _urlquote, unquote as urlunquote
from urllib.parse import urlparse, urlsplit, urlunsplit, urljoin
from http.cookiejar import MozillaCookieJar


Kilobyte = 1024
Megabyte = Kilobyte ** 2

# Set the default chunk size above the minimum 5MB chunk size for AWS S3 multipart uploads, but also try to find a sweet
# spot between a minimal number of chunks and payloads big enough to be efficient but not incur retries for slow clients
# or situations where network connectivity is unreliable for whatever reason.
DEFAULT_CHUNK_SIZE = Megabyte * 25
# 5GB is the max chunk limit imposed by AWS S3, also a (generous) hard limit.
HARD_MAX_CHUNK_SIZE = Megabyte * 1000 * 5
# Max number of "parts" for AWS S3.
DEFAULT_MAX_CHUNK_LIMIT = 10000
# A practical default limit for single request body payload size, similar to AWS S3 recommendation for payload sizes.
DEFAULT_MAX_REQUEST_SIZE = Megabyte * 100

DEFAULT_HEADERS = {}
DEFAULT_CONFIG_PATH = os.path.join(os.path.expanduser('~'), '.deriva')
DEFAULT_CREDENTIAL_FILE = os.path.join(DEFAULT_CONFIG_PATH, 'credential.json')
DEFAULT_GLOBUS_CREDENTIAL_FILE = os.path.join(DEFAULT_CONFIG_PATH, 'globus-credential.json')
DEFAULT_CONFIG_FILE = os.path.join(DEFAULT_CONFIG_PATH, 'config.json')
DEFAULT_COOKIE_JAR_FILE = os.path.join(DEFAULT_CONFIG_PATH, 'cookies.txt')
DEFAULT_REQUESTS_TIMEOUT = (6, 63)  # (connect, read), integer in seconds
DEFAULT_SESSION_CONFIG = {
    "timeout": DEFAULT_REQUESTS_TIMEOUT,
    "retry_connect": 2,
    "retry_read": 4,
    "retry_backoff_factor": 1.0,
    "retry_status_forcelist": [500, 502, 503, 504],
    "allow_retry_on_all_methods": False,
    "cookie_jar": DEFAULT_COOKIE_JAR_FILE,
    "max_request_size": DEFAULT_MAX_REQUEST_SIZE,
    "max_chunk_limit": DEFAULT_MAX_CHUNK_LIMIT
}
OAUTH2_SCOPES_KEY = "oauth2_scopes"
DEFAULT_CONFIG = {
    "server":
    {
        "protocol": "https",
        "host": 'localhost',
        "catalog_id": 1
    },
    "session": DEFAULT_SESSION_CONFIG,
    "download_processor_whitelist": [],
    OAUTH2_SCOPES_KEY: {}
}
DEFAULT_CREDENTIAL = {}
DEFAULT_LOGGER_OVERRIDES = {
    "globus_sdk": logging.WARNING,
    "boto3": logging.WARNING,
    "botocore": logging.WARNING,
}


class NotModified (ValueError):
    pass


class ConcurrentUpdate (ValueError):
    pass


def urlquote(s, safe=''):
    """Quote all reserved characters according to RFC3986 unless told otherwise.

       The urllib.urlquote has a weird default which excludes '/' from
       quoting even though it is a reserved character.  We would never
       want this when encoding elements in Deriva REST API URLs, so
       this wrapper changes the default to have no declared safe
       characters.

    """
    return _urlquote(s.encode('utf-8'), safe=safe)


def urlquote_dcctx(s, safe='~{}",:'):
    """Quote for use with Deriva-Client-Context or other HTTP headers.

       Defaults to allow additional safe characters for less
       aggressive encoding of JSON content for use in an HTTP header
       value.

    """
    return urlquote(s, safe=safe)


def stob(string):
    return bool(du_util.strtobool(str(string)))


def format_exception(e):
    if not isinstance(e, Exception):
        return str(e)
    exc = "".join(("[", type(e).__name__, "] "))
    if isinstance(e, requests.HTTPError):
        resp = " - Server responded: %s" % e.response.text.strip().replace('\n', ': ') if e.response.text else ""
        return "".join((exc, str(e), resp))
    return "".join((exc, str(e)))


def add_logging_level(level_name, level_num, method_name=None):
    if not method_name:
        method_name = level_name.lower()

    if hasattr(logging, level_name):
        logging.warning('{} already defined in logging module'.format(level_name))
        return
    if hasattr(logging, method_name):
        logging.warning('{} already defined in logging module'.format(method_name))
        return
    if hasattr(logging.getLoggerClass(), method_name):
        logging.warning('{} already defined in logger class'.format(method_name))
        return

    def log_for_level(self, message, *args, **kwargs):
        if self.isEnabledFor(level_num):
            self._log(level_num, message, args, **kwargs)

    def log_to_root(message, *args, **kwargs):
        logging.log(level_num, message, *args, **kwargs)

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, log_for_level)
    setattr(logging, method_name, log_to_root)


def init_logging(level=logging.INFO,
                 log_format=None,
                 file_path=None,
                 file_mode='w',
                 capture_warnings=True,
                 logger_config=DEFAULT_LOGGER_OVERRIDES):
    add_logging_level("TRACE", logging.DEBUG-5)
    logging.captureWarnings(capture_warnings)
    if log_format is None:
        log_format = "[%(asctime)s - %(levelname)s - %(name)s:%(filename)s:%(lineno)s:%(funcName)s()] %(message)s" \
            if level <= logging.DEBUG else "%(asctime)s - %(levelname)s - %(message)s"
    # allow for reconfiguration of module-specific logging levels
    [logging.getLogger(name).setLevel(level) for name, level in logger_config.items()]
    if file_path:
        logging.basicConfig(filename=file_path, filemode=file_mode, level=level, format=log_format)
    else:
        logging.basicConfig(level=level, format=log_format)


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        self.timeout = DEFAULT_REQUESTS_TIMEOUT
        if "timeout" in kwargs:
            timeout = kwargs["timeout"]
            self.timeout = tuple(timeout) if isinstance(timeout, list) else timeout
            del kwargs["timeout"]
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


def get_new_requests_session(url=None, session_config=DEFAULT_SESSION_CONFIG):
    session = requests.session()
    retries = Retry(connect=session_config['retry_connect'],
                    read=session_config['retry_read'],
                    backoff_factor=session_config['retry_backoff_factor'],
                    status_forcelist=session_config['retry_status_forcelist'],
                    allowed_methods=Retry.DEFAULT_ALLOWED_METHODS if
                    # Passing False to method_whitelist means allow all methods
                    not session_config.get("allow_retry_on_all_methods", False) else False,
                    raise_on_status=True)
    adapter = TimeoutHTTPAdapter(timeout=session_config.get("timeout", DEFAULT_REQUESTS_TIMEOUT), max_retries=retries)
    if url:
        session.mount(url, adapter)
    else:
        session.mount('http://', adapter)
        session.mount('https://', adapter)

    return session


def make_dirs(path, mode=0o777):
    if not os.path.isdir(path):
        try:
            os.makedirs(path, mode=mode)
        except OSError as error:
            if error.errno != errno.EEXIST:
                raise


def copy_config(src, dst):
    config_dir = os.path.dirname(dst)
    make_dirs(config_dir, mode=0o750)
    shutil.copy2(src, dst)


def write_config(config_file=DEFAULT_CONFIG_FILE, config=DEFAULT_CONFIG):
    config_dir = os.path.dirname(config_file)
    make_dirs(config_dir, mode=0o750)
    with io.open(config_file, 'w', newline='\n', encoding='utf-8') as cf:
        config_data = json.dumps(config, ensure_ascii=False, indent=2)
        cf.write(config_data)
        cf.close()


def read_config(config_file=DEFAULT_CONFIG_FILE, create_default=False, default=DEFAULT_CONFIG):
    if not config_file:
        config_file = DEFAULT_CONFIG_FILE
    config = None
    if not os.path.isfile(config_file) and create_default:
        logging.info("No default configuration file found, attempting to create one at: %s" % config_file)
        try:
            write_config(config_file, default)
        except Exception as e:
            logging.warning("Unable to create configuration file %s. Using internal defaults. %s" %
                            (config_file, format_exception(e)))
            config = json.dumps(default, ensure_ascii=False)

    if not config:
        with open(config_file, encoding='utf-8') as cf:
            config = cf.read()

    return json.loads(config, object_pairs_hook=OrderedDict)


def lock_file(file_path, mode, exclusive=True, timeout=60):
    return portalocker.Lock(file_path, mode=mode, timeout=timeout, fail_when_locked=True,
                            flags=(portalocker.LOCK_EX | portalocker.LOCK_NB) if exclusive else
                            (portalocker.LOCK_SH | portalocker.LOCK_NB))


def write_credential(credential_file=DEFAULT_CREDENTIAL_FILE, credential=DEFAULT_CREDENTIAL):
    credential_dir = os.path.dirname(credential_file)
    make_dirs(credential_dir, mode=0o750)
    with lock_file(credential_file, mode='w', exclusive=True) as cf:
        os.chmod(credential_file, 0o600)
        credential_data = json.dumps(credential, ensure_ascii=False, indent=2)
        cf.write(credential_data)
        cf.flush()
        os.fsync(cf.fileno())


def read_credential(credential_file=DEFAULT_CREDENTIAL_FILE, create_default=False, default=DEFAULT_CREDENTIAL):
    if not credential_file:
        credential_file = DEFAULT_CREDENTIAL_FILE
    credential = None
    if not os.path.isfile(credential_file) and create_default:
        logging.info("No default credential file found, attempting to create one at: %s" % credential_file)
        try:
            write_credential(credential_file, default)
        except Exception as e:
            logging.warning("Unable to create credential file %s. Using internal defaults. %s" %
                            (credential_file, format_exception(e)))
            credential = json.dumps(default, ensure_ascii=False)

    if not credential:
        with lock_file(credential_file, mode='r', exclusive=False) as cf:
            credential = cf.read()

    return json.loads(credential, object_pairs_hook=OrderedDict)


def get_oauth_scopes_for_host(host,
                              config_file=DEFAULT_CONFIG_FILE,
                              force_refresh=False,
                              warn_on_discovery_failure=False):
    config = read_config(config_file or DEFAULT_CONFIG_FILE, create_default=True)
    required_scopes = config.get(OAUTH2_SCOPES_KEY)
    result = dict()
    upr = urlparse(host)
    if upr.scheme and upr.netloc:
        if upr.scheme not in ("http", "https"):
            return result
        url = urljoin(host, "/authn/discovery")
        host = upr.hostname
    else:
        url = "https://%s/authn/discovery" % host
    # determine the scope to use based on host-to-scope(s) mappings in the config file
    if required_scopes:
        for hostname, scopes in required_scopes.items():
            if host.lower() == hostname.lower():
                result = scopes
                break
    if not result or force_refresh:
        session = get_new_requests_session(session_config=DEFAULT_SESSION_CONFIG)
        try:
            r = session.get(url, headers=DEFAULT_HEADERS)
            r.raise_for_status()
            result = r.json().get(OAUTH2_SCOPES_KEY)
            if result:
                if not config.get(OAUTH2_SCOPES_KEY):
                    config[OAUTH2_SCOPES_KEY] = {}
                config[OAUTH2_SCOPES_KEY].update({host: result})
                write_config(config_file or DEFAULT_CONFIG_FILE, config=config)
        except Exception as e:
            msg = "Unable to discover and/or update the \"%s\" mappings from [%s]. Requests to this host that " \
                  "use OAuth2 scope-specific bearer tokens may fail or provide only limited access. %s" % \
                  (OAUTH2_SCOPES_KEY, url, format_exception(e))
            if warn_on_discovery_failure:
                logging.warning(msg)
            else:
                logging.debug(msg)
        finally:
            session.close()
    return result


def format_credential(token=None, oauth2_token=None, username=None, password=None):
    if username and password:
        return {"username": username, "password": password}
    credential = dict()
    if token:
        credential.update({"cookie": "webauthn=%s" % token})
    if oauth2_token:
        credential.update({"bearer-token": "%s" % oauth2_token})
    if not credential:
        raise ValueError(
            "Missing required argument(s): a supported authentication token or a username/password must be provided.")

    return credential


def bootstrap(logging_level=logging.INFO):
    init_logging(level=logging_level)
    read_config(create_default=True)
    read_credential(create_default=True)


def load_cookies_from_file(cookie_file=None):
    if not cookie_file:
        cookie_file = DEFAULT_SESSION_CONFIG["cookie_jar"]
    cookies = MozillaCookieJar()
    # Load and return saved cookies if existing
    if os.path.isfile(cookie_file):
        try:
            cookies.load(cookie_file, ignore_discard=True, ignore_expires=True)
            return cookies
        except Exception as e:
            logging.warning(format_exception(e))

    # Create new empty cookie file otherwise
    cookies.save(cookie_file, ignore_discard=True, ignore_expires=True)
    os.chmod(cookie_file, 0o600)

    return cookies


def resource_path(relative_path, default=os.path.abspath(".")):
    if default is None:
        return relative_path
    return os.path.join(default, relative_path)


def get_transfer_summary(total_bytes, elapsed_time):
    total_secs = elapsed_time.total_seconds()
    transferred = \
        float(total_bytes) / float(Kilobyte) if total_bytes < Megabyte else float(total_bytes) / float(Megabyte)
    throughput = str(" at %.2f MB/second" % (transferred / total_secs)) if (total_secs >= 1) else ""
    elapsed = str("Elapsed time: %s." % elapsed_time) if (total_secs > 0) else ""
    summary = "%.2f %s transferred%s. %s" % \
              (transferred, "KB" if total_bytes < Megabyte else "MB", throughput, elapsed)
    return summary


def calculate_optimal_transfer_shape(size,
                                     chunk_limit=DEFAULT_MAX_CHUNK_LIMIT,
                                     requested_chunk_size=DEFAULT_CHUNK_SIZE,
                                     byte_align=Kilobyte * 64):
    if size == 0:
        return DEFAULT_CHUNK_SIZE, 1, 0
    if size < 0:
        raise ValueError('Parameter "size" cannot be negative')
    if chunk_limit <= 0:
        raise ValueError('Parameter "chunk_limit" must be greater than zero')
    if byte_align <= 0:
        raise ValueError('Parameter "byte_align" must be greater than zero')

    if size > (HARD_MAX_CHUNK_SIZE * chunk_limit):
        raise ValueError("Size %d exceeds limit for max chunk size %d and chunk count %d" %
                         (size, HARD_MAX_CHUNK_SIZE, chunk_limit))
    calculated_chunk_size = math.ceil(math.ceil(size / chunk_limit) / byte_align) * byte_align
    chosen_chunk_size = min(HARD_MAX_CHUNK_SIZE, max(requested_chunk_size, calculated_chunk_size))
    remainder = size % chosen_chunk_size
    chunk_count = size // chosen_chunk_size
    if remainder:
        chunk_count += 1
    return chosen_chunk_size, chunk_count, remainder


def json_item_handler(input_file, callback):
    with io.open(input_file, "r", encoding='utf-8') as infile:
        line = infile.readline().lstrip()
        infile.seek(0)
        is_json_stream = False
        if line.startswith('{') and line.endswith('}\n'):
            input_json = infile
            is_json_stream = True
        else:
            input_json = json.load(infile, object_pairs_hook=OrderedDict)
        try:
            for item in input_json:
                if is_json_stream:
                    item = json.loads(item, object_pairs_hook=OrderedDict)
                if callback:
                    callback(item)
        finally:
            infile.close()


def topo_sorted(depmap):
    """Return list of items topologically sorted.

       depmap: { item: [required_item, ...], ... }

    Raises ValueError if a required_item cannot be satisfied in any order.

    The per-item required_item iterables must allow revisiting on
    multiple iterations.

    """
    ordered = [ item for item, requires in depmap.items() if not requires ]
    depmap = { item: set(requires) for item, requires in depmap.items() if requires }
    satisfied = set(ordered)
    while depmap:
        additions = []
        for item, requires in list(depmap.items()):
            if requires.issubset(satisfied):
                additions.append(item)
                satisfied.add(item)
                del depmap[item]
        if not additions:
            raise ValueError(("unsatisfiable", depmap))
        ordered.extend(additions)
        additions = []
    return ordered


class AttrDict (dict):
    """Dictionary with optional attribute-based lookup.

       For keys that are valid attributes, self.key is equivalent to
       self[key].
    """
    def __getattr__(self, a):
        try:
            return self[a]
        except KeyError as e:
            raise AttributeError(str(e))

    def __setattr__(self, a, v):
        self[a] = v

    def update(self, d):
        dict.update(self, d)


# convenient enumeration of common annotation tags
tag = AttrDict({
    'display':            'tag:misd.isi.edu,2015:display',
    'table_alternatives': 'tag:isrd.isi.edu,2016:table-alternatives',
    'column_display':     'tag:isrd.isi.edu,2016:column-display',
    'key_display':        'tag:isrd.isi.edu,2017:key-display',
    'foreign_key':        'tag:isrd.isi.edu,2016:foreign-key',
    'generated':          'tag:isrd.isi.edu,2016:generated',
    'immutable':          'tag:isrd.isi.edu,2016:immutable',
    'non_deletable':      'tag:isrd.isi.edu,2016:non-deletable',
    'app_links':          'tag:isrd.isi.edu,2016:app-links',
    'table_display':      'tag:isrd.isi.edu,2016:table-display',
    'visible_columns':    'tag:isrd.isi.edu,2016:visible-columns',
    'visible_foreign_keys': 'tag:isrd.isi.edu,2016:visible-foreign-keys',
    'export':             'tag:isrd.isi.edu,2016:export',
    'export_2019':        'tag:isrd.isi.edu,2019:export',
    'export_fragment_definitions': 'tag:isrd.isi.edu,2021:export-fragment-definitions',    
    'asset':              'tag:isrd.isi.edu,2017:asset',
    'citation':           'tag:isrd.isi.edu,2018:citation',
    'required':           'tag:isrd.isi.edu,2018:required',
    'indexing_preferences': 'tag:isrd.isi.edu,2018:indexing-preferences',
    'bulk_upload':        'tag:isrd.isi.edu,2017:bulk-upload',
    'chaise_config':      'tag:isrd.isi.edu,2019:chaise-config',
    'source_definitions': 'tag:isrd.isi.edu,2019:source-definitions',
    'indexing_preferences': 'tag:isrd.isi.edu,2018:indexing-preferences',
    'google_dataset':     'tag:isrd.isi.edu,2021:google-dataset',
    'column_defaults':    'tag:isrd.isi.edu,2023:column-defaults',
    'viz_3d_display':     'tag:isrd.isi.edu,2021:viz-3d-display',
})



import logging
import os
import sys
import requests
import traceback
from pprint import pp
from requests.exceptions import HTTPError, ConnectionError
from deriva.core import __version__ as VERSION, BaseCLI, KeyValuePairArgs, DerivaServer, DerivaPathError, \
    get_credential, format_credential, format_exception, DEFAULT_HEADERS
from deriva.core.ermrest_model import nochange
from deriva.core.utils import eprint


class DerivaCatalogCLIException (Exception):
    """Base exception class for DerivaHatracCli.
    """
    def __init__(self, message):
        """Initializes the exception.
        """
        super(DerivaCatalogCLIException, self).__init__(message)


class UsageException (DerivaCatalogCLIException):
    """Usage exception.
    """
    def __init__(self, message):
        """Initializes the exception.
        """
        super(UsageException, self).__init__(message)


class ResourceException (DerivaCatalogCLIException):
    """Remote resource exception.
    """
    def __init__(self, message, cause):
        """Initializes the exception.
        """
        super(ResourceException, self).__init__(message)
        self.cause = cause


class DerivaCatalogCLI (BaseCLI):
    """Deriva Hatrac Command-line Interface.
    """
    def __init__(self, description, epilog):
        """Initializes the CLI.
        """
        super(DerivaCatalogCLI, self).__init__(description, epilog, VERSION)

        # initialized after argument parsing
        self.args = None
        self.host = None
        self.protocol = None
        self.catalog = None

        # parent arg parser
        self.remove_options(['--config-file', '--credential-file'])
        self.parser.add_argument("--protocol", choices=["http", "https"], default='https',
                                 help="transport protocol: 'http' or 'https'")
        subparsers = self.parser.add_subparsers(title='sub-commands', dest='subcmd')

        # exists parser
        exists_parser = subparsers.add_parser('exists', help="Check if catalog exists.")
        exists_parser.add_argument("id", metavar="<id>", type=str, help="catalog id")
        exists_parser.set_defaults(func=self.catalog_exists)

        # create parser
        create_parser = subparsers.add_parser('create', help="Create a new catalog.")
        create_parser.add_argument("--id", metavar="<id>", type=str, help="catalog id")
        create_parser.add_argument("--owner", metavar="<owner> <owner> ...",
                                   nargs="+", help="List of quoted user or group identifier strings.")
        create_parser.add_argument("--auto-configure", action="store_true",
                                   help="Configure the new catalog with a set of baseline defaults")
        create_parser.add_argument("--configure-args", metavar="[key=value key=value ...]",
                                   nargs='+', action=KeyValuePairArgs, default={},
                                   help="Variable length of whitespace-delimited key=value pair arguments used for "
                                        "controlling automatic catalog configuration settings. "
                                        "For example: --configure-args key1=value1 key2=value2")
        create_parser.set_defaults(func=self.catalog_create)

        # get parser
        get_parser = subparsers.add_parser('get', help="Send a HTTP GET request to the catalog.")
        get_parser.add_argument("id", metavar="<id>", type=str, help="catalog id")
        get_parser.add_argument("path", metavar="<request-path>", help="The ERMRest API path.")
        get_parser.add_argument("--output-file", metavar="<output file path>", help="Path to output file.")
        get_parser.add_argument("--output-format", choices=["json", "json-stream", "csv"], default="json")
        get_parser.add_argument("--auto-delete", action="store_true",
                                help="Automatically delete output file if no results are returned.")
        get_parser.add_argument("--headers", metavar="[key=value key=value ...]",
                                nargs='+', action=KeyValuePairArgs, default={},
                                help="Variable length of whitespace-delimited key=value pair arguments used for "
                                     "specifying or overriding HTTP Request Headers. "
                                     "For example: --headers key1=value1 key2=value2")
        get_parser.set_defaults(func=self.catalog_get)

        # put parser
        put_parser = subparsers.add_parser('put', help="Send a HTTP PUT request to the catalog.")
        put_parser.add_argument("id", metavar="<id>", type=str, help="catalog id")
        put_parser.add_argument("path", metavar="<request-path>", help="The ERMRest API path.")
        put_parser.add_argument("--input-file", metavar="<output file path>", help="Path to input file.")
        put_parser.add_argument("--input-format", choices=["json", "json-stream", "csv"], default="json")
        put_parser.add_argument("--headers", metavar="[key=value key=value ...]",
                                nargs='+', action=KeyValuePairArgs, default={},
                                help="Variable length of whitespace-delimited key=value pair arguments used for "
                                     "specifying or overriding HTTP Request Headers. "
                                     "For example: --headers key1=value1 key2=value2")
        put_parser.set_defaults(func=self.catalog_put)

        # post parser
        post_parser = subparsers.add_parser('post', help="Send a HTTP POST request to the catalog.")
        post_parser.add_argument("id", metavar="<id>", type=str, help="catalog id")
        post_parser.add_argument("path", metavar="<request-path>", help="The ERMRest API path.")
        post_parser.add_argument("--input-file", metavar="<input file path>", help="Path to input file.")
        post_parser.add_argument("--input-format", choices=["json", "json-stream", "csv"], default="json")
        post_parser.add_argument("--headers", metavar="[key=value key=value ...]",
                                 nargs='+', action=KeyValuePairArgs, default={},
                                 help="Variable length of whitespace-delimited key=value pair arguments used for "
                                      "specifying or overriding HTTP Request Headers. "
                                      "For example: --headers key1=value1 key2=value2")
        post_parser.set_defaults(func=self.catalog_post)

        # delete parser
        del_parser = subparsers.add_parser('delete', help="Send a HTTP DELETE request to the catalog. "
                                                          "Use the 'drop' command to delete the entire catalog.")
        del_parser.add_argument("id", metavar="<id>", type=str, help="catalog id")
        del_parser.add_argument("path", metavar="<request-path>", help="The ERMRest API path.")
        del_parser.add_argument("--headers", metavar="[key=value key=value ...]",
                                nargs='+', action=KeyValuePairArgs, default={},
                                help="Variable length of whitespace-delimited key=value pair arguments used for "
                                     "specifying or overriding HTTP Request Headers. "
                                     "For example: --headers key1=value1 key2=value2")
        del_parser.set_defaults(func=self.catalog_delete)

        # drop parser
        drop_parser = subparsers.add_parser('drop', help="Delete a catalog.")
        drop_parser.add_argument("id", metavar="<id>", type=str, help="catalog id")
        drop_parser.set_defaults(func=self.catalog_drop)

        # clone parser
        clone_parser = subparsers.add_parser('clone', help="Clone a source catalog to a new destination catalog.")
        clone_parser.add_argument("id", metavar="<id>", type=str, help="catalog id")
        clone_parser.add_argument("--no-copy-data", action="store_false",
                                  help="Do not copy table contents.")
        clone_parser.add_argument("--no-copy-annotations", action="store_false",
                                  help="Do not copy annotations.")
        clone_parser.add_argument("--no-copy-policy", action="store_false",
                                  help="Do not copy ACL (Access Control List) policies.")
        clone_parser.add_argument("--no-truncate-after", action="store_false",
                                  help="Do not truncate destination history after cloning.")
        clone_parser.add_argument("--exclude-schemas", metavar="<schema-name> <schema-name> ...",
                                  nargs="+", help="List of schema names to exclude from the cloning process.")
        clone_parser.set_defaults(func=self.catalog_clone)

        # create_alias parser
        create_alias_parser = subparsers.add_parser('create-alias', help="Create a new catalog alias")
        create_alias_parser.add_argument("--id", metavar="<id>", type=str, help="The alias id.")
        create_alias_parser.add_argument("--alias-target", metavar="<alias>", help="The target catalog id.")
        create_alias_parser.add_argument("--owner", metavar="<owner> <owner> ...",
                                         nargs="+", help="List of quoted user or group identifier strings.")
        create_alias_parser.set_defaults(func=self.catalog_alias_create)

        # get_alias parser
        get_alias_parser = subparsers.add_parser('get-alias', help="Get catalog alias metadata")
        get_alias_parser.add_argument("id", metavar="<id>", type=str, help="The alias id.")
        get_alias_parser.set_defaults(func=self.catalog_alias_get)

        # update_alias parser
        update_alias_parser = subparsers.add_parser('update-alias', help="Update an existing catalog alias")
        update_alias_parser.add_argument("--id", metavar="<id>", type=str, help="The alias id.")
        update_alias_parser.add_argument("--alias-target", metavar="<alias>", nargs='?', default=nochange, const=None,
                                         help="The target catalog id. If specified without a catalog id as an argument "
                                              "value, the existing alias target will be cleared ")
        update_alias_parser.add_argument("--owner", metavar="<owner> <owner> ...", nargs='+', default=nochange,
                                         help="List of quoted user or group identifier strings.")
        update_alias_parser.set_defaults(func=self.catalog_alias_update)

        # delete_alias parser
        del_alias_parser = subparsers.add_parser('delete-alias', help="Delete a catalog alias.")
        del_alias_parser.add_argument("id", metavar="<id>", type=str, help="The alias id.")
        del_alias_parser.set_defaults(func=self.catalog_alias_delete)

    @staticmethod
    def _get_credential(host_name, token=None, oauth2_token=None):
        if token or oauth2_token:
            return format_credential(token=token, oauth2_token=oauth2_token)
        else:
            return get_credential(host_name)

    def _post_parser_init(self, args):
        """Shared initialization for all sub-commands.
        """
        self.host = args.host if args.host else 'localhost'
        self.protocol = args.protocol
        self.id = args.id
        self.server = DerivaServer(self.protocol,
                                   args.host,
                                   credentials=DerivaCatalogCLI._get_credential(self.host,
                                                                                token=args.token,
                                                                                oauth2_token=args.oauth2_token))

    def catalog_exists(self, args):
        """Implements the catalog_exists sub-command.
        """
        catalog = self.server.connect_ermrest(self.id)
        pp(catalog.exists())

    def catalog_create(self, args):
        """Implements the catalog_create sub-command.
        """
        try:
            if args.id and self.server.connect_ermrest(args.id).exists():
                print("Catalog already exists")
                return
            owner = args.owner if args.owner else None
            catalog = self.server.create_ermrest_catalog(args.id, owner)
            if args.auto_configure:
                model = catalog.getCatalogModel()
                model.configure_baseline_catalog(**args.configure_args)
            if not args.quiet:
                print("Created new catalog %s with the following default configuration:\n" % catalog.catalog_id)
                pp(catalog.get('/').json())
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException('Catalog not found', e)
            elif e.response.status_code == requests.codes.conflict:
                raise ResourceException("Catalog already exists", e)
            else:
                raise e

    def catalog_clone(self, args):
        """Implements the catalog_clone sub-command.
        """
        try:
            catalog = self.server.connect_ermrest(args.id)
            print("Attempting to clone catalog %s into new catalog. Please wait..." % args.id)
            dest_cat = catalog.clone_catalog(copy_data=args.no_copy_data,
                                             copy_annotations=args.no_copy_annotations,
                                             copy_policy=args.no_copy_policy,
                                             truncate_after=args.no_truncate_after,
                                             exclude_schemas=args.exclude_schemas)
            print("Catalog successfully cloned into new catalog: %s" % dest_cat.catalog_id)
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException('Catalog not found', e)
            else:
                raise e

    def catalog_get(self, args):
        """Implements the catalog_get sub-command.
        """
        headers = DEFAULT_HEADERS.copy()
        headers.update(args.headers)

        if args.output_format == "json":
            headers["accept"] = "application/json"
        elif args.output_format == "json-stream":
            headers["accept"] = "application/x-json-stream"
        elif args.output_format == "csv":
            headers["accept"] = "text/csv"
        else:
            raise UsageException("Unsupported output format: %s" % args.output_format)

        catalog = self.server.connect_ermrest(args.id)
        try:
            if args.output_file:
                catalog.getAsFile(args.path,
                                  destfilename=args.output_file,
                                  headers=headers,
                                  delete_if_empty=args.auto_delete)
            else:
                pp(catalog.get(args.path, headers=headers).json())
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException('Catalog not found', e)
        except:
            if args.output_file and os.path.isfile(args.output_file):
                logging.info("Deleting empty file: %s" % args.output_file)
                os.remove(args.output_file)
            raise

    # TODO: implement PUT at some point
    def catalog_put(self, args):
        raise NotImplementedError

    # TODO: implement POST at some point
    def catalog_post(self, args):
        raise NotImplementedError

    def catalog_delete(self, args):
        """Implements the catalog_delete sub-command.
        """
        headers = DEFAULT_HEADERS.copy()
        headers.update(args.headers)
        try:
            catalog = self.server.connect_ermrest(args.id)
            catalog.delete(args.path, headers)
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException('Catalog not found', e)
            else:
                raise e

    def catalog_drop(self, args):
        """Implements the catalog_drop sub-command.
        """
        try:
            catalog = self.server.connect_ermrest(args.id)
            catalog.delete_ermrest_catalog(really=True)
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException('Catalog not found', e)
            else:
                raise e

    def catalog_alias_create(self, args):
        """Implements the catalog_alias_create sub-command.
        """
        try:
            if args.id:
                alias = self.server.connect_ermrest_alias(args.id)
                try:
                    if alias.retrieve():
                        print("Catalog alias already exists")
                        return
                except requests.HTTPError as e:
                    if e.response.status_code == 404:
                        pass
                    else:
                        raise
            owner = args.owner if args.owner else None
            alias = self.server.create_ermrest_alias(args.id, owner, args.alias_target)
            if not args.quiet:
                print("Created new catalog alias %s with the following configuration:\n" % alias.alias_id)
                pp(alias.retrieve())
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException('Catalog alias not found', e)
            elif e.response.status_code == requests.codes.conflict:
                raise ResourceException("Catalog alias already exists", e)
            else:
                raise

    def catalog_alias_get(self, args):
        """Implements the catalog_alias_get sub-command.
        """
        try:
            alias = self.server.connect_ermrest_alias(args.id)
            response = alias.retrieve()
            if not args.quiet:
                pp(response)
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException('Catalog alias not found', e)
            else:
                raise e

    def catalog_alias_update(self, args):
        try:
            owner = args.owner if args.owner else None
            alias = self.server.connect_ermrest_alias(args.id)
            response = alias.update(owner, args.alias_target, args.id)
            print("Updated catalog alias %s with the following configuration:\n" % alias.alias_id)
            pp(response)
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException('Catalog alias not found', e)
            elif e.response.status_code == requests.codes.conflict:
                raise ResourceException("Catalog alias already exists", e)
            else:
                raise

    def catalog_alias_delete(self, args):
        """Implements the catalog_alias_delete sub-command.
        """
        try:
            alias = self.server.connect_ermrest_alias(args.id)
            alias.delete_ermrest_alias(really=True)
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException('Catalog alias not found', e)
            else:
                raise e

    def main(self):
        """Main routine of the CLI.
        """
        args = self.parse_cli()

        def _resource_error_message(emsg):
            return "{prog} {subcmd}: {id}: {msg}".format(
                prog=self.parser.prog, subcmd=args.subcmd, id=args.id, msg=emsg)

        try:
            if not hasattr(args, 'func'):
                self.parser.print_usage()
                return 1

            self._post_parser_init(args)
            args.func(args)
            return 0
        except UsageException as e:
            eprint("{prog} {subcmd}: {msg}".format(prog=self.parser.prog, subcmd=args.subcmd, msg=e))
        except ConnectionError as e:
            eprint("{prog}: Connection error occurred".format(prog=self.parser.prog))
        except DerivaPathError as e:
            eprint(e)
        except HTTPError as e:
            if e.response.status_code == requests.codes.unauthorized:
                msg = 'Authentication required'
            elif e.response.status_code == requests.codes.forbidden:
                msg = 'Permission denied'
            else:
                msg = e
            logging.debug(format_exception(e))
            eprint(_resource_error_message(msg))
        except ResourceException as e:
            logging.debug(format_exception(e.cause))
            eprint(_resource_error_message(e))
        except RuntimeError as e:
            logging.warning(format_exception(e))
            eprint('Unexpected runtime error occurred')
        except:
            eprint('Unexpected error occurred')
            traceback.print_exc()
        return 1


def main():
    DESC = "DERIVA Catalog Utility Command-Line Interface"
    INFO = "For more information see: https://github.com/informatics-isi-edu/deriva-py"
    return DerivaCatalogCLI(DESC, INFO).main()


if __name__ == '__main__':
    sys.exit(main())



import os
import sys
import json
import traceback
import argparse
import requests
from requests.exceptions import HTTPError, ConnectionError
from deriva.core import BaseCLI, KeyValuePairArgs, format_credential, format_exception, urlparse
from deriva.transfer import DerivaRestore, DerivaRestoreError, DerivaRestoreConfigurationError, \
    DerivaRestoreAuthenticationError, DerivaRestoreAuthorizationError


class DerivaRestoreCLI(BaseCLI):
    def __init__(self, description, epilog, **kwargs):

        BaseCLI.__init__(self, description, epilog, **kwargs)
        self.parser.add_argument("--catalog", metavar="<1>", help="Catalog number. If a catalog number is not "
                                                                  "specified, a new catalog will be created.")
        self.parser.add_argument("input_path", metavar="<input_path>", help="Path to backup file or directory.")
        mutex_group = self.parser.add_mutually_exclusive_group()
        mutex_group.add_argument("--no-data", action="store_true",
                                 help="Do not restore table data, restore schema only.")
        mutex_group.add_argument("--no-schema", action="store_true",
                                 help="Do not restore schema, restore data only.")
        self.parser.add_argument("--no-assets", action="store_true",
                                 help="Do not restore asset data, if present.")
        self.parser.add_argument("--no-annotations", action="store_true",
                                 help="Do not restore annotations.")
        self.parser.add_argument("--no-policy", action="store_true",
                                 help="Do not restore access policy and ACLs.")
        self.parser.add_argument("--no-bag-materialize", action="store_true",
                                 help="If the input format is a bag, do not materialize prior to restore.")
        self.parser.add_argument("--weak-bag-validation", action="store_true",
                                 help="If the input format is a bag, "
                                      "do not abort the restore if the bag fails validation.")
        self.parser.add_argument("--exclude-object", type=lambda s: [item.strip() for item in s.split(',')],
                                 metavar="<schema>, <schema:table>, ...",
                                 help="List of comma-delimited schema-name and/or schema-name/table-name to "
                                      "exclude from the restore process, in the form <schema> or <schema:table>.")
        self.parser.add_argument("--exclude-data", type=lambda s: [item.strip() for item in s.split(',')],
                                 metavar="<schema>, <schema:table>, ...",
                                 help="List of comma-delimited schema-name and/or schema-name/table-name to "
                                      "exclude from the restore process, in the form <schema> or <schema:table>.")
        self.parser.add_argument("envars", metavar="[key=value key=value ...]",
                                 nargs=argparse.REMAINDER, action=KeyValuePairArgs, default={},
                                 help="Variable length of whitespace-delimited key=value pair arguments used for "
                                      "populating the processing environment with parameters for keyword substitution."
                                      "For example: key1=value1 key2=value2")

    def main(self):
        try:
            args = self.parse_cli()
        except ValueError as e:
            sys.stderr.write(str(e))
            return 2
        if not args.quiet:
            sys.stderr.write("\n")

        try:
            assert args.host, "A hostname is required!"
            server = dict()
            server["catalog_id"] = args.catalog
            if args.host.startswith("http"):
                url = urlparse(args.host)
                server["protocol"] = url.scheme
                server["host"] = url.netloc
            else:
                server["protocol"] = "https"
                server["host"] = args.host

            restorer = DerivaRestore(server, **vars(args), dcctx_cid="cli/" + self.__class__.__name__)
            try:
                restorer.restore()
            except ConnectionError as e:
                raise DerivaRestoreError("Connection error occurred. %s" % format_exception(e))
            except HTTPError as e:
                if e.response.status_code == requests.codes.unauthorized:
                    raise DerivaRestoreAuthenticationError(
                        "The requested service requires authentication and a valid login session could "
                        "not be found for the specified host. Server responded: %s" % e)
                elif e.response.status_code == requests.codes.forbidden:
                    raise DerivaRestoreAuthorizationError(
                        "A requested operation was forbidden. Server responded: %s" % e)
                raise DerivaRestoreError(format_exception(e))
        except (DerivaRestoreError, DerivaRestoreConfigurationError,
                DerivaRestoreAuthenticationError, DerivaRestoreAuthorizationError) as e:
            sys.stderr.write(("\n" if not args.quiet else "") + format_exception(e))
            if args.debug:
                traceback.print_exc()
            return 1
        except:
            sys.stderr.write("An unexpected error occurred.")
            traceback.print_exc()
            return 1
        finally:
            if not args.quiet:
                sys.stderr.write("\n\n")
        return 0



import os
import time
import datetime
import uuid
import logging
import platform
import requests
from requests.exceptions import HTTPError
from bdbag import bdbag_api as bdb, bdbag_ro as ro, BAG_PROFILE_TAG, BDBAG_RO_PROFILE_ID
from deriva.core import ErmrestCatalog, HatracStore, format_exception, get_credential, format_credential, read_config, \
    stob, Megabyte, __version__ as VERSION
from deriva.core.utils.version_utils import get_installed_version
from deriva.transfer.download.processors import find_query_processor, find_transform_processor, find_post_processor
from deriva.transfer.download.processors.base_processor import LOCAL_PATH_KEY, REMOTE_PATHS_KEY, SERVICE_URL_KEY, \
    FILE_SIZE_KEY
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError, \
    DerivaDownloadAuthenticationError, DerivaDownloadAuthorizationError, DerivaDownloadTimeoutError


class DerivaDownload(object):
    """

    """
    def __init__(self, server, **kwargs):
        self.server = server
        self.hostname = None
        self.catalog = None
        self.store = None
        self.cancelled = False
        self.output_dir = os.path.abspath(kwargs.get("output_dir", "."))
        self.envars = kwargs.get("envars", dict())
        self.config = kwargs.get("config")
        self.credentials = kwargs.get("credentials", dict())
        config_file = kwargs.get("config_file")
        credential_file = kwargs.get("credential_file")
        self.metadata = dict()
        self.sessions = dict()
        self.allow_anonymous = kwargs.get("allow_anonymous", True)
        self.max_payload_size_mb = int(kwargs.get("max_payload_size_mb", 0) or 0)
        self.timeout_secs = int(kwargs.get("timeout", 0) or 0)
        self.timeout = None

        info = "%s v%s [Python %s, %s]" % (
            self.__class__.__name__, get_installed_version(VERSION),
            platform.python_version(), platform.platform(aliased=True))
        logging.info("Initializing downloader: %s" % info)

        if not self.server:
            raise DerivaDownloadConfigurationError("Server not specified!")

        # server variable initialization
        self.hostname = self.server.get('host', '')
        if not self.hostname:
            raise DerivaDownloadConfigurationError("Host not specified!")
        protocol = self.server.get('protocol', 'https')
        self.server_url = protocol + "://" + self.hostname
        catalog_id = self.server.get("catalog_id", "1")
        session_config = self.server.get('session')

        # credential initialization
        token = kwargs.get("token")
        oauth2_token = kwargs.get("oauth2_token")
        username = kwargs.get("username")
        password = kwargs.get("password")
        if credential_file:
            self.credentials = get_credential(self.hostname, credential_file)
        elif token or oauth2_token or (username and password):
            self.credentials = format_credential(token=token,
                                                 oauth2_token=oauth2_token,
                                                 username=username,
                                                 password=password)

        # catalog and file store initialization
        if self.catalog:
            del self.catalog
        self.catalog = ErmrestCatalog(
            protocol, self.hostname, catalog_id, self.credentials, session_config=session_config)
        if self.store:
            del self.store
        self.store = HatracStore(
            protocol, self.hostname, self.credentials, session_config=session_config)

        # init dcctx cid
        self.set_dcctx_cid(kwargs.get("dcctx_cid", "api/" + self.__class__.__name__))

        # process config file
        if config_file:
            try:
                self.config = read_config(config_file)
            except Exception as e:
                raise DerivaDownloadConfigurationError(e)

    def set_dcctx_cid(self, cid):
        assert cid, "A dcctx cid is required"
        if self.catalog:
            self.catalog.dcctx['cid'] = cid
        if self.store:
            self.store.dcctx['cid'] = cid

    def set_config(self, config):
        self.config = config

    def set_credentials(self, credentials):
        self.catalog.set_credentials(credentials, self.hostname)
        self.store.set_credentials(credentials, self.hostname)
        self.credentials = credentials

    def check_payload_size(self, outputs):
        if self.max_payload_size_mb < 1:
            return

        # This is not very well optimized for the way it is invoked but until it becomes an issue, it is good enough.
        max_bytes = self.max_payload_size_mb * Megabyte
        total_bytes = 0
        for v in outputs.values():
            total_bytes += v.get(FILE_SIZE_KEY, 0)
            if total_bytes >= max_bytes:
                raise DerivaDownloadError("Maximum payload size of %d megabytes exceeded." %
                                          self.max_payload_size_mb)

    def download(self, **kwargs):

        if not self.config:
            raise DerivaDownloadConfigurationError("No configuration specified!")

        if self.config.get("catalog") is None:
            raise DerivaDownloadConfigurationError("Catalog configuration error!")

        if self.timeout_secs > 0:
            self.timeout = datetime.datetime.now() + datetime.timedelta(0, self.timeout_secs)

        ro_manifest = None
        ro_author_name = None
        ro_author_orcid = None
        remote_file_manifest = os.path.abspath(
            ''.join([os.path.join(self.output_dir, 'remote-file-manifest_'), str(uuid.uuid4()), ".json"]))

        catalog_config = self.config['catalog']
        self.envars.update(self.config.get('env', dict()))
        self.envars.update({"hostname": self.hostname})

        # 1. If we don't have a client identity, we need to authenticate
        identity = kwargs.get("identity")
        if not identity:
            try:
                if not self.credentials:
                    self.set_credentials(get_credential(self.hostname))
                logging.info("Validating credentials for host: %s" % self.hostname)
                attributes = self.catalog.get_authn_session().json()
                identity = attributes["client"]
            except HTTPError as he:
                if he.response.status_code == 404:
                    logging.info("No existing login session found for host: %s" % self.hostname)
            except Exception as e:
                raise DerivaDownloadAuthenticationError("Unable to validate credentials: %s" % format_exception(e))
        wallet = kwargs.get("wallet", {})

        # 2. Check for bagging config and initialize bag related variables
        bag_path = None
        bag_archiver = None
        bag_algorithms = None
        bag_idempotent = False
        bag_config = self.config.get('bag')
        create_bag = True if bag_config else False
        if create_bag:
            bag_name = bag_config.get(
                'bag_name', ''.join(["deriva_bag", '_', time.strftime("%Y-%m-%d_%H.%M.%S")])).format(**self.envars)
            bag_path = os.path.abspath(os.path.join(self.output_dir, bag_name))
            bag_archiver = bag_config.get('bag_archiver')
            bag_algorithms = bag_config.get('bag_algorithms', ['sha256'])
            bag_idempotent = stob(bag_config.get('bag_idempotent', False))
            bag_metadata = bag_config.get('bag_metadata', {"Internal-Sender-Identifier":
                                                           "deriva@%s" % self.server_url})
            bag_ro = create_bag and not bag_idempotent and stob(bag_config.get('bag_ro', "True"))
            if create_bag:
                bdb.ensure_bag_path_exists(bag_path)
                bag = bdb.make_bag(bag_path, algs=bag_algorithms, metadata=bag_metadata, idempotent=bag_idempotent)
                if bag_ro:
                    ro_author_name = bag.info.get("Contact-Name",
                                                  None if not identity else
                                                  identity.get('full_name',
                                                               identity.get('display_name',
                                                                            identity.get('id', None))))
                    ro_author_orcid = bag.info.get("Contact-Orcid")
                    ro_manifest = ro.init_ro_manifest(author_name=ro_author_name, author_orcid=ro_author_orcid)
                    bag_metadata.update({BAG_PROFILE_TAG: BDBAG_RO_PROFILE_ID})

        # 3. Process the set of queries by locating, instantiating, and invoking the specified processor(s)
        outputs = dict()
        base_path = bag_path if bag_path else self.output_dir
        for processor in catalog_config['query_processors']:
            processor_name = processor["processor"]
            processor_type = processor.get('processor_type')
            processor_params = processor.get('processor_params')

            try:
                query_processor = find_query_processor(processor_name, processor_type)
                processor = query_processor(self.envars,
                                            inputs=outputs,
                                            bag=create_bag,
                                            catalog=self.catalog,
                                            store=self.store,
                                            base_path=base_path,
                                            processor_params=processor_params,
                                            remote_file_manifest=remote_file_manifest,
                                            ro_manifest=ro_manifest,
                                            ro_author_name=ro_author_name,
                                            ro_author_orcid=ro_author_orcid,
                                            identity=identity,
                                            wallet=wallet,
                                            allow_anonymous=self.allow_anonymous,
                                            timeout=self.timeout)
                outputs = processor.process()
                if processor.should_abort():
                    raise DerivaDownloadTimeoutError("Timeout (%s seconds) waiting for processor [%s] to complete." %
                                                     (self.timeout_secs, processor_name))
                self.check_payload_size(outputs)
            except Exception as e:
                logging.error(format_exception(e))
                if create_bag:
                    bdb.cleanup_bag(bag_path)
                    if remote_file_manifest and os.path.isfile(remote_file_manifest):
                        os.remove(remote_file_manifest)
                raise

        # 4. Execute anything in the transform processing pipeline, if configured
        transform_processors = self.config.get('transform_processors', [])
        if transform_processors:
            for processor in transform_processors:
                processor_name = processor["processor"]
                processor_type = processor.get('processor_type')
                processor_params = processor.get('processor_params')
                try:
                    transform_processor = find_transform_processor(processor_name, processor_type)
                    processor = transform_processor(
                        self.envars,
                        inputs=outputs,
                        processor_params=processor_params,
                        base_path=base_path,
                        bag=create_bag,
                        ro_manifest=ro_manifest,
                        ro_author_name=ro_author_name,
                        ro_author_orcid=ro_author_orcid,
                        identity=identity,
                        wallet=wallet,
                        allow_anonymous=self.allow_anonymous,
                        timeout=self.timeout)
                    outputs = processor.process()
                    if processor.should_abort():
                        raise DerivaDownloadTimeoutError(
                            "Timeout (%s seconds) waiting for processor [%s] to complete." %
                            (self.timeout_secs, processor_name))
                    self.check_payload_size(outputs)
                except Exception as e:
                    if create_bag:
                        bdb.cleanup_bag(bag_path)
                        if remote_file_manifest and os.path.isfile(remote_file_manifest):
                            os.remove(remote_file_manifest)
                    raise

        # 5. Create the bag, and archive (serialize) if necessary
        if create_bag:
            try:
                if ro_manifest:
                    ro.write_bag_ro_metadata(ro_manifest, bag_path)
                if not os.path.isfile(remote_file_manifest):
                    remote_file_manifest = None
                bdb.make_bag(bag_path,
                             algs=bag_algorithms,
                             remote_file_manifest=remote_file_manifest
                             if (remote_file_manifest and os.path.getsize(remote_file_manifest) > 0) else None,
                             update=True,
                             idempotent=bag_idempotent)
            except Exception as e:
                logging.fatal("Exception while updating bag manifests: %s" % format_exception(e))
                bdb.cleanup_bag(bag_path)
                raise
            finally:
                if remote_file_manifest and os.path.isfile(remote_file_manifest):
                    os.remove(remote_file_manifest)

            logging.info('Created bag: %s' % bag_path)

            if bag_archiver is not None:
                try:
                    archive = bdb.archive_bag(bag_path,
                                              bag_archiver.lower(),
                                              idempotent=bag_idempotent)
                    bdb.cleanup_bag(bag_path)
                    outputs = {os.path.basename(archive): {LOCAL_PATH_KEY: archive}}
                except Exception as e:
                    logging.error("Exception while creating data bag archive: %s" % format_exception(e))
                    raise
            else:
                outputs = {os.path.basename(bag_path): {LOCAL_PATH_KEY: bag_path}}

        # 6. Execute anything in the post processing pipeline, if configured
        post_processors = self.config.get('post_processors', [])
        if post_processors:
            for processor in post_processors:
                processor_name = processor["processor"]
                processor_type = processor.get('processor_type')
                processor_params = processor.get('processor_params')
                try:
                    post_processor = find_post_processor(processor_name, processor_type)
                    processor = post_processor(
                        self.envars,
                        inputs=outputs,
                        processor_params=processor_params,
                        identity=identity,
                        wallet=wallet,
                        allow_anonymous=self.allow_anonymous,
                        timeout=self.timeout)
                    outputs = processor.process()
                    if processor.should_abort():
                        raise DerivaDownloadTimeoutError(
                            "Timeout (%s seconds) waiting for processor [%s] to complete." %
                            (self.timeout_secs, processor_name))
                    self.check_payload_size(outputs)
                except Exception as e:
                    logging.error(format_exception(e))
                    raise

        return outputs

    def __del__(self):
        for session in self.sessions.values():
            session.close()


class GenericDownloader(DerivaDownload):
    LOCAL_PATH_KEY = LOCAL_PATH_KEY
    REMOTE_PATHS_KEY = REMOTE_PATHS_KEY
    SERVICE_URL_KEY = SERVICE_URL_KEY

    def __init__(self, *args, **kwargs):
        DerivaDownload.__init__(self, *args, **kwargs)



import json
import os
import re
from string import Template
import logging
from deriva.core import make_dirs
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError
from deriva.transfer.download.processors.transform.base_transform_processor import BaseTransformProcessor

logger = logging.getLogger(__name__)

INPUT_PATHS_KEY = "input_paths"


class InterpolationTransformProcessor(BaseTransformProcessor):
    """String interpolation transform processor.
    """
    def __init__(self, envars=None, **kwargs):
        super(InterpolationTransformProcessor, self).__init__(envars, **kwargs)
        self._create_input_output_paths()
        # get custom param
        self.template = Template(self.parameters.get('template'))
        logger.debug("Interpolating with template: {}".format(self.template.template))

    def process(self):
        """Reads a json-stream input file, performs string interpolation, and writes to output file.
        """
        try:
            with open(self.input_abspath, encoding='utf-8') as inputfile, \
                 open(self.output_abspath, mode='w', encoding='utf-8') as outputfile:
                for line in inputfile:
                    row = json.loads(line)
                    output = self.template.safe_substitute(row)
                    outputfile.write(output)
        except IOError as e:
            raise DerivaDownloadError("Interpolation transform failed", e)

        return super(InterpolationTransformProcessor, self).process()


class StrSubTransformProcessor(BaseTransformProcessor):
    """String substitution transform processor.
    """
    def __init__(self, envars=None, **kwargs):
        super(StrSubTransformProcessor, self).__init__(envars, **kwargs)
        self._create_input_output_paths()
        # get custom param
        self.substitutions = self.parameters.get('substitutions', [])
        logger.debug("String substitutions: {}".format(self.substitutions))
        # validate strsubs
        for strsub in self.substitutions:
            missing = set(strsub.keys()) - {'pattern', 'repl', 'input', 'output'}
            if missing:
                raise DerivaDownloadConfigurationError("Missing required key(s) %s in 'substitutions' parameter" % missing)

    def process(self):
        """Reads a json-stream input file, performs string substitutions, and writes to output file.
        """
        try:
            with open(self.input_abspath, encoding='utf-8') as inputfile, \
                 open(self.output_abspath, mode='w', encoding='utf-8') as outputfile:
                for line in inputfile:
                    row = json.loads(line)
                    for strsub in self.substitutions:
                        row[strsub['output']] = re.sub(strsub['pattern'], strsub['repl'], row[strsub['input']])
                    outputfile.write(json.dumps(row))
                    outputfile.write('\n')
        except IOError as e:
            raise DerivaDownloadError("Interpolation transform failed", e)
        except KeyError as e:
            raise DerivaDownloadError("Required input attribute not found in row", e)

        return super(StrSubTransformProcessor, self).process()


class ConcatenateTransformProcessor(BaseTransformProcessor):
    """Concatenate transform processor.
    """
    def __init__(self, envars=None, **kwargs):
        super(ConcatenateTransformProcessor, self).__init__(envars, **kwargs)
        self._create_input_output_paths()

    def process(self):
        """Reads a input files in order given and writes to output file.
        """
        try:
            make_dirs(os.path.dirname(self.output_abspath))
            with open(self.output_abspath, mode='w', encoding='utf-8') as outputfile:
                for input_abspath in self.input_abspaths:
                    with open(input_abspath, encoding='utf-8') as inputfile:
                        for line in inputfile:
                            outputfile.write(line)
        except IOError as e:
            raise DerivaDownloadError("Concatenate transform failed", e)

        return super(ConcatenateTransformProcessor, self).process()



import logging
import os
from os.path import basename
import requests
from requests.exceptions import HTTPError, ConnectionError
import sys
import traceback
from deriva.core import __version__ as VERSION, BaseCLI, DerivaPathError, HatracStore, HatracHashMismatch, \
    get_credential, format_credential, format_exception, DEFAULT_CHUNK_SIZE
from deriva.core.utils import eprint, mime_utils as mu


class DerivaHatracCLIException (Exception):
    """Base exception class for DerivaHatracCli.
    """
    def __init__(self, message):
        """Initializes the exception.
        """
        super(DerivaHatracCLIException, self).__init__(message)


class UsageException (DerivaHatracCLIException):
    """Usage exception.
    """
    def __init__(self, message):
        """Initializes the exception.
        """
        super(UsageException, self).__init__(message)


class ResourceException (DerivaHatracCLIException):
    """Remote resource exception.
    """
    def __init__(self, message, cause):
        """Initializes the exception.
        """
        super(ResourceException, self).__init__(message)
        self.cause = cause


class DerivaHatracCLI (BaseCLI):
    """Deriva Hatrac Command-line Interface.
    """
    def __init__(self, description, epilog):
        """Initializes the CLI.
        """
        super(DerivaHatracCLI, self).__init__(description, epilog, VERSION)

        # initialized after argument parsing
        self.args = None
        self.host = None
        self.resource = None
        self.store = None

        # parent arg parser
        self.remove_options(['--config-file', '--credential-file'])
        subparsers = self.parser.add_subparsers(title='sub-commands', dest='subcmd')

        # list parser
        ls_parser = subparsers.add_parser('ls', help="list the elements of a namespace")
        ls_parser.add_argument("resource", metavar="<path>", type=str, help="namespace path")
        ls_parser.set_defaults(func=self.list)

        # mkdir parser
        mkdir_parser = subparsers.add_parser('mkdir', help="create a namespace")
        mkdir_parser.add_argument("-p", "--parents", action="store_true",
                                  help="Create intermediate parent namespaces as required")
        mkdir_parser.add_argument("resource", metavar="<path>", type=str, help="namespace path")
        mkdir_parser.set_defaults(func=self.mkdir)

        # rmdir parser
        rmdir_parser = subparsers.add_parser('rmdir', help="remove a namespace")
        rmdir_parser.add_argument("resource", metavar="<path>", type=str, help="namespace path")
        rmdir_parser.set_defaults(func=self.rmdir)

        # getacl parser
        getacl_parser = subparsers.add_parser('getacl', help="get ACL")
        getacl_parser.add_argument("resource", metavar="<path>", type=str, help="object or namespace path")
        getacl_parser.add_argument("--access", default=None, metavar="<access-mode>",
                                   help="Optionally specify 'access' mode.")
        getacl_parser.add_argument("--role", default=None, metavar="<role>",
                                   help="Optionally specify 'role'. Must specify 'access' with this option.")
        getacl_parser.set_defaults(func=self.getacl)

        # setacl parser
        setacl_parser = subparsers.add_parser('setacl', help="set ACL")
        setacl_parser.add_argument("resource", metavar="<path>", type=str, help="object or namespace path")
        setacl_parser.add_argument("access", metavar="<access-mode>", help="access mode")
        setacl_parser.add_argument("roles", nargs='+', metavar="<role>", help="role")
        setacl_parser.add_argument("--add", action="store_true", help="add a single role to the ACL")
        setacl_parser.set_defaults(func=self.setacl)

        # detacl parser
        delacl_parser = subparsers.add_parser('delacl', help="delete ACL")
        delacl_parser.add_argument("resource", metavar="<path>", type=str, help="object or namespace path")
        delacl_parser.add_argument("access", metavar="<access-mode>", help="access mode")
        delacl_parser.add_argument("role", nargs='?', metavar="<role>", help="role")
        delacl_parser.set_defaults(func=self.delacl)

        # getobj parser
        getobj_parser = subparsers.add_parser('get', help="get object")
        getobj_parser.add_argument("resource", metavar="<path>", type=str, help="object path")
        getobj_parser.add_argument('outfile', metavar="<outfile>", nargs='?', type=str, help="output filename or -")
        getobj_parser.set_defaults(func=self.getobj)

        # putobj parser
        putobj_parser = subparsers.add_parser('put', help="put object")
        putobj_parser.add_argument('infile', metavar="<infile>", type=str, help="input filename")
        putobj_parser.add_argument("resource", metavar="<path>", type=str, help="object path")
        putobj_parser.add_argument("--content-type", metavar="<type>", type=str, help="HTTP Content-Type header value")
        putobj_parser.add_argument("--chunk-size", metavar="<bytes>", type=int, default=DEFAULT_CHUNK_SIZE,
                                   help="Chunk size in bytes")
        putobj_parser.add_argument("--parents", action="store_true",
                                   help="Create intermediate parent namespaces as required")
        putobj_parser.set_defaults(func=self.putobj)

        # delobj parser
        delobj_parser = subparsers.add_parser('del', help="delete object")
        delobj_parser.add_argument("resource", metavar="<path>", type=str, help="object path")
        delobj_parser.set_defaults(func=self.delobj)

    @staticmethod
    def _get_credential(host_name, token=None, oauth2_token=None):
        if token or oauth2_token:
            return format_credential(token=token, oauth2_token=oauth2_token)
        else:
            return get_credential(host_name)

    def _post_parser_init(self, args):
        """Shared initialization for all sub-commands.
        """
        self.host = args.host if args.host else 'localhost'
        self.resource = args.resource
        self.store = HatracStore('https', args.host, DerivaHatracCLI._get_credential(self.host,
                                                                                     token=args.token,
                                                                                     oauth2_token=args.oauth2_token))

    def list(self, args):
        """Implements the list sub-command.
        """
        try:
            namespaces = self.store.retrieve_namespace(self.resource)
            for name in namespaces:
                print(name)
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException('No such object or namespace', e)
            elif e.response.status_code != requests.codes.conflict:
                # 'conflict' just means the namespace has no contents - ok
                raise e
        except ValueError as e:
            raise ResourceException('Not a namespace', e)

    def mkdir(self, args):
        """Implements the mkdir sub-command.
        """
        try:
            self.store.create_namespace(self.resource, parents=args.parents)
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException("Parent namespace not found (use '--parents' to create parent namespace)", e)
            elif e.response.status_code == requests.codes.conflict:
                raise ResourceException("Namespace exists or the parent path is not a namespace", e)
            else:
                raise e

    def rmdir(self, args):
        """Implements the mkdir sub-command.
        """
        try:
            self.store.delete_namespace(self.resource)
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException('No such object or namespace', e)
            elif e.response.status_code == requests.codes.conflict:
                raise ResourceException("Namespace not empty", e)
            else:
                raise e

    def getacl(self, args):
        """Implements the getacl sub-command.
        """
        if args.role and not args.access:
            raise UsageException("Must use '--access' option with '--role' option")

        try:
            acls = self.store.get_acl(self.resource, args.access, args.role)
            for access in acls:
                print("%s:" % access)
                for role in acls.get(access, []):
                    print("  %s" % role)
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException('No such object or namespace or ACL entry', e)
            elif e.response.status_code == requests.codes.bad_request:
                raise ResourceException('Invalid ACL name %s' % args.access, e)
            else:
                raise e

    def setacl(self, args):
        """Implements the setacl sub-command.
        """
        if args.add and len(args.roles) > 1:
            raise UsageException("Option '--add' is only valid for a single role")

        try:
            self.store.set_acl(self.resource, args.access, args.roles, args.add)
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException('No such object or namespace', e)
            elif e.response.status_code == requests.codes.bad_request:
                raise ResourceException('Resource cannot be updated as requested', e)
            else:
                raise e

    def delacl(self, args):
        """Implements the getacl sub-command.
        """
        try:
            self.store.del_acl(self.resource, args.access, args.role)
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException('No such object or namespace or ACL entry', e)
            elif e.response.status_code == requests.codes.bad_request:
                raise ResourceException('Resource cannot be updated as requested', e)
            else:
                raise e

    def getobj(self, args):
        """Implements the getobj sub-command.
        """
        try:
            if args.outfile and args.outfile == '-':
                r = self.store.get_obj(self.resource)
                logging.debug('Content encoding: %s' % r.apparent_encoding)
                assert r.content, 'content cannot be read as bytes'  # never expected from the requests API
                os.write(sys.stdout.fileno(), r.content)
            else:
                outfilename = args.outfile if args.outfile else basename(self.resource)
                self.store.get_obj(self.resource, destfilename=outfilename)
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException('No such object', e)
            else:
                raise e

    def putobj(self, args):
        """Implements the putobj sub-command.
        """
        try:
            content_type = args.content_type if args.content_type else mu.guess_content_type(args.infile)
            file_size = os.path.getsize(args.infile)
            loc = self.store.put_loc(
                self.resource,
                args.infile,
                content_type=content_type,
                chunked=True if file_size > args.chunk_size else False,
                chunk_size=args.chunk_size,
                create_parents=args.parents)
            print(loc)
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException("Parent namespace not found (use '--parents' to create parent namespace)", e)
            elif e.response.status_code == requests.codes.conflict:
                raise ResourceException(
                    'Cannot create object (parent path is not a namespace or object name is in use)', e)
            else:
                raise e

    def delobj(self, args):
        """Implements the delobj sub-command.
        """
        try:
            self.store.del_obj(self.resource)
        except HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise ResourceException('No such object', e)
            else:
                raise e

    def main(self):
        """Main routine of the CLI.
        """
        args = self.parse_cli()

        def _resource_error_message(emsg):
            return "{prog} {subcmd}: {resource}: {msg}".format(
                prog=self.parser.prog, subcmd=args.subcmd, resource=args.resource, msg=emsg)

        try:
            if not hasattr(args, 'func'):
                self.parser.print_usage()
                return 1

            self._post_parser_init(args)
            args.func(args)
            return 0
        except UsageException as e:
            eprint("{prog} {subcmd}: {msg}".format(prog=self.parser.prog, subcmd=args.subcmd, msg=e))
        except ConnectionError as e:
            eprint("{prog}: Connection error occurred".format(prog=self.parser.prog))
        except DerivaPathError as e:
            eprint(e)
        except HTTPError as e:
            if e.response.status_code == requests.codes.unauthorized:
                msg = 'Authentication required'
            elif e.response.status_code == requests.codes.forbidden:
                msg = 'Permission denied'
            else:
                msg = e
            logging.debug(format_exception(e))
            eprint(_resource_error_message(msg))
        except ResourceException as e:
            logging.debug(format_exception(e.cause))
            eprint(_resource_error_message(e))
        except HatracHashMismatch as e:
            logging.debug(format_exception(e))
            eprint(_resource_error_message('Checksum verification failed'))
        except RuntimeError as e:
            logging.debug(format_exception(e))
            eprint('Unexpected runtime error occurred')
        except:
            eprint('Unexpected error occurred')
            traceback.print_exc()
        return 1


def main():
    DESC = "DERIVA HATRAC Command-Line Interface"
    INFO = "For more information see: https://github.com/informatics-isi-edu/deriva-py"
    return DerivaHatracCLI(DESC, INFO).main()


if __name__ == '__main__':
    sys.exit(main())



import io
import os
import json
import logging
from enum import Enum
from deriva.core import json_item_handler
from deriva.transfer.download import DerivaDownloadError, DerivaDownloadConfigurationError
from deriva.transfer.download.processors.transform.base_transform_processor import BaseTransformProcessor

logger = logging.getLogger(__name__)


ColumnFunctions = Enum("ColumnFunctions", ["add", "replace", "delete"])


class ColumnTransformProcessor(BaseTransformProcessor):
    """Column transform processor.
    """
    def __init__(self, envars=None, **kwargs):
        super(ColumnTransformProcessor, self).__init__(envars, **kwargs)
        self._create_input_output_paths()
        self.ro_file_provenance = False
        # get custom params
        self.column_transforms = self.parameters.get("column_transforms", dict())
        self.writer = None

    def process(self):
        """Reads a json-stream input file, performs column manipulation functions, and writes to output file.
        """
        try:
            with open(self.output_abspath, mode='w', encoding='utf-8') as output_file:
                def row_handler(item):
                    if not self.writer:
                        self.writer = io.open(self.output_abspath, 'w', encoding='utf-8')
                    for k, v in self.column_transforms.items():
                        func = v["fn"]
                        value = v.get("value")
                        if func == ColumnFunctions.add.name:
                            item.update({k: value})
                        elif func == ColumnFunctions.delete.name:
                            del item[k]
                        elif func == ColumnFunctions.replace.name:
                            repl_k = value["key"]
                            repl_v = value.get("value")
                            if repl_v:
                                item.update({k: item[repl_k][repl_v]})
                            else:
                                item.update({k: item[repl_k]})
                        else:
                            raise DerivaDownloadError(
                                "Unknown function '%s' in column transform [%s]" % (func, v))
                    output_file.writelines(''.join(
                        [json.dumps(item, ensure_ascii=False), '\n']))
                json_item_handler(self.input_abspath, row_handler)
                self.writer.flush()
                self.writer.close()

        except IOError as e:
            raise DerivaDownloadError("Column transform failed", e)

        return super(ColumnTransformProcessor, self).process()



from deriva.transfer.download.deriva_download import DerivaDownload, GenericDownloader, DerivaDownloadError, \
    DerivaDownloadConfigurationError, DerivaDownloadAuthenticationError, DerivaDownloadAuthorizationError
from deriva.transfer.download.deriva_download_cli import DerivaDownloadCLI

from deriva.transfer.upload.deriva_upload import DerivaUpload, GenericUploader, DerivaUploadError, DerivaUploadError, \
    DerivaUploadConfigurationError, DerivaUploadCatalogCreateError, DerivaUploadCatalogUpdateError
from deriva.transfer.upload.deriva_upload_cli import DerivaUploadCLI

from deriva.transfer.backup.deriva_backup import DerivaBackup, DerivaBackupAuthenticationError, \
    DerivaBackupAuthorizationError, DerivaBackupConfigurationError, DerivaBackupError
from deriva.transfer.backup.deriva_backup_cli import DerivaBackupCLI

from deriva.transfer.restore.deriva_restore import DerivaRestore, DerivaRestoreAuthenticationError, \
    DerivaRestoreAuthorizationError, DerivaRestoreConfigurationError, DerivaRestoreError
from deriva.transfer.restore.deriva_restore_cli import DerivaRestoreCLI



import sys
import json
import re
from deriva.core import ErmrestCatalog, AttrDict, ermrest_model, get_credential, __version__ as VERSION, \
    format_exception, urlquote
from deriva.config.base_config import BaseSpec, BaseSpecList, ConfigUtil, ConfigBaseCLI
from requests.exceptions import HTTPError
from uuid import UUID
import warnings


class NoForeignKeyError(ValueError):
    pass


class ACLSpecList(BaseSpecList):
    def __init__(self, dictlist=None):
        BaseSpecList.__init__(self, ACLSpec, dictlist)


class ACLSpec(BaseSpec):
    def __init__(self, specdict):
        BaseSpec.__init__(self, specdict, ["acl", "no_acl", "acl_bindings", "invalidate_bindings"], "acl")

    def validate(self):
        BaseSpec.validate(self)
        if self.get("no_acl") not in [True, False, None]:
            raise ValueError("no_acl must be True or False (or not present)")
        if self.get("acl") is not None and self.get("no_acl"):
            raise ValueError("can't specify an acl and no_acl=True in the same spec")
        if self.get("acl") is None and self.get("no_acl") == False:
            raise ValueError("if no_acl=False, an acl must be specified")


class AclConfig:
    NC_NAME = 'name'
    GC_NAME = 'groups'
    ACL_TYPES = ["catalog_acl", "schema_acls", "table_acls", "column_acls", "foreign_key_acls"]
    GLOBUS_PREFIX = 'https://auth.globus.org/'
    ROBOT_PREFIX_FORMAT = 'https://{server}/webauthn_robot/'

    def __init__(self, server, catalog_id, config_file, credentials, schema_name=None, table_name=None, verbose=False):
        self.config = json.load(open(config_file))
        self.ignored_schema_patterns = []
        self.verbose = verbose
        self.server = server
        self.catalog_id = catalog_id
        ip = self.config.get("ignored_schema_patterns")
        if ip is not None:
            for p in ip:
                self.ignored_schema_patterns.append(re.compile(p))
        self.acl_specs = {"catalog_acl": self.config.get("catalog_acl")}
        for key in self.ACL_TYPES:
            if key != "catalog_acl":
                self.acl_specs[key] = self.make_speclist(key)
        self.groups = self.config.get("groups")
        self.expand_groups()
        self.acl_definitions = self.config.get("acl_definitions")
        self.expand_acl_definitions()
        self.acl_bindings = self.config.get("acl_bindings")
        self.invalidate_bindings = self.config.get("invalidate_bindings")

        old_catalog = ErmrestCatalog('https', self.server, self.catalog_id, credentials)
        self.saved_toplevel_config = ConfigUtil.find_toplevel_node(old_catalog.getCatalogModel(), schema_name,
                                                                   table_name)
        self.catalog = ErmrestCatalog('https', self.server, self.catalog_id, credentials)
        self.toplevel_config = ConfigUtil.find_toplevel_node(self.catalog.getCatalogModel(), schema_name, table_name)

    def make_speclist(self, name):
        d = self.config.get(name)
        if d is None:
            d = dict()
        return ACLSpecList(d)

    def add_node_acl(self, node, acl_name):
        acl = self.acl_definitions.get(acl_name)
        if acl is None:
            raise ValueError("no acl set called '{name}'".format(name=acl_name))
        for k in acl.keys():
            node.acls[k] = acl[k]

    def add_node_acl_binding(self, node, table_node, binding_name):
        if not binding_name in self.acl_bindings:
            raise ValueError("no acl binding called '{name}'".format(name=binding_name))
        binding = self.acl_bindings.get(binding_name)
        try:
            node.acl_bindings[binding_name] = self.expand_acl_binding(binding, table_node)
        except NoForeignKeyError as e:
            detail = ''
            if isinstance(node, ermrest_model.Column):
                detail = 'on column {n}'.format(n=node.name)
            elif isinstance(node, ermrest_model.ForeignKey):
                detail = 'on foreign key {s}.{n}'.format(s=node.names[0][0], n=node.names[0][1])
            else:
                detail = ' {t}'.format(t=type(node))
            print(
                "couldn't expand acl binding {b} {d} table {s}.{t}".format(b=binding_name, d=detail, s=table_node.schema.name,
                                                                           t=table_node.name))
            raise e

    def expand_acl_binding(self, binding, table_node):
        if not isinstance(binding, dict):
            return binding
        new_binding = dict()
        for k in binding.keys():
            if k == "projection":
                new_binding[k] = []
                for proj in binding.get(k):
                    new_binding[k].append(self.expand_projection(proj, table_node))
            elif k == "scope_acl":
                new_binding[k] = self.get_group(binding.get(k))
            else:
                new_binding[k] = binding[k]
        return new_binding

    def expand_projection(self, proj, table_node):
        if isinstance(proj, dict):
            new_proj = dict()
            is_first_outbound = True
            for k in proj.keys():
                if k == "outbound_col":
                    if is_first_outbound:
                        is_first_outbound = False
                    else:
                        raise NotImplementedError(
                            "don't know how to expand 'outbound_col' on anything but the first entry in a projection; "
                            "use 'outbound' instead")
                    if table_node is None:
                        raise NotImplementedError(
                            "don't know how to expand 'outbound_col' in a foreign key acl/annotation; use 'outbound' "
                            "instead")
                    new_proj["outbound"] = self.expand_projection_column(proj[k], table_node)
                    if new_proj["outbound"] is None:
                        return None
                else:
                    new_proj[k] = proj[k]
                    is_first_outbound = False
            return new_proj
        else:
            return proj

    def expand_projection_column(self, col_name, table_node):
        for fkey in table_node.foreign_keys:
            if len(fkey.foreign_key_columns) == 1:
                col = fkey.foreign_key_columns[0]
                if col.table.name == table_node.name and col.table.schema.name == table_node.schema.name and col.name == col_name:
                    return fkey.names[0]
        raise NoForeignKeyError("can't find foreign key for column %I.%I(%I)", table_node.schema.name, table_node.name,
                                col_name)

    def set_node_acl_bindings(self, node, table_node, binding_list, invalidate_list):
        node.acl_bindings.clear()
        if binding_list is not None:
            for binding_name in binding_list:
                self.add_node_acl_binding(node, table_node, binding_name)
        if invalidate_list is not None:
            for binding_name in invalidate_list:
                if binding_list and binding_name in binding_list:
                    raise ValueError(
                        "Binding {b} appears in both acl_bindings and invalidate_bindings for table {s}.{t} node {n}".format(
                            b=binding_name, s=table_node.schema.name, t=table_node.name, n=node.name))
                node.acl_bindings[binding_name] = False

    def save_groups(self):
        glt = self.create_or_validate_group_table()
        if glt is not None and self.groups is not None:
            rows = []
            for name in self.groups.keys():
                row = {'name': name, 'groups': self.groups.get(name)}
                for c in ['RCB', 'RMB']:
                    if glt.getColumn(c) is not None:
                        row[c] = None
                rows.append(row)

            glt.upsertRows(self.catalog, rows)

    def create_or_validate_schema(self, schema_name):
        schema = self.catalog.getCatalogSchema()['schemas'].get(schema_name)
        if schema is None:
            self.catalog.post("/schema/{s}".format(s=schema_name))
        return self.catalog.getCatalogSchema()['schemas'].get(schema_name)

    def create_table(self, schema_name, table_name, table_spec, comment=None):
        if table_spec is None:
            table_spec = dict()
        if schema_name is None:
            return None
        table_spec["schema_name"] = schema_name
        table_spec["table_name"] = table_name
        if table_spec.get('comment') is None and comment is not None:
            table_spec['comment'] = comment
        if table_spec.get('kind') is None:
            table_spec['kind'] = 'table'
        self.catalog.post("/schema/{s}/table".format(s=schema_name), json=table_spec)
        schema = self.catalog.getCatalogSchema()['schemas'].get(schema_name)
        return schema['tables'].get(table_name)

    def create_or_validate_group_table(self):
        glt_spec = self.config.get('group_list_table')
        if glt_spec is None:
            return None
        sname = glt_spec.get('schema')
        tname = glt_spec.get('table')
        if sname is None or tname is None:
            raise ValueError("group_list_table missing schema or table")
        schema = self.create_or_validate_schema(sname)
        assert schema is not None
        glt = Table(schema['tables'].get(tname))
        if glt == {}:
            glt_spec = ermrest_model.Table.define(
                tname,
                column_defs=[
                    ermrest_model.Column.define(
                        self.NC_NAME,
                        ermrest_model.builtin_types.text,
                        nullok=False,
                        comment='Name of grouplist, used in foreign keys. This table is maintained by the acl-config '
                                'program and should not be updated by hand.'
                    ),
                    ermrest_model.Column.define(
                        self.GC_NAME,
                        ermrest_model.builtin_types['text[]'],
                        nullok=True,
                        comment='List of groups. This table is maintained by the acl-config program and should not be '
                                'updated by hand.'
                    )
                ],
                key_defs=[
                    ermrest_model.Key.define(
                        [self.NC_NAME],
                        constraint_names=[[sname, "{t}_{c}_u".format(t=tname, c=self.NC_NAME)]]
                    )
                ],
                comment="Named lists of groups used in ACLs. Maintained by the acl-config program. Do not update this "
                        "table manually.",
                annotations={'tag:isrd.isi.edu,2016:generated': None}
            )
            glt = Table(self.create_table(sname, tname, glt_spec))

        else:
            name_col = glt.getColumn(self.NC_NAME)
            if name_col is None:
                raise ValueError(
                    'table specified for group lists ({s}.{t}) lacks a "{n}" column'.format(s=sname, t=tname,
                                                                                            n=self.NC_NAME))
            if name_col.get('nullok'):
                raise ValueError(
                    "{n} column in group list table ({s}.{t}) allows nulls".format(n=self.NC_NAME, s=sname, t=tname))

            nc_uniq = False
            for key in glt.get('keys'):
                cols = key.get('unique_columns')
                if len(cols) == 1 and cols[0] == self.NC_NAME:
                    nc_uniq = True
                    break
            if not nc_uniq:
                raise ValueError(
                    "{n} column in group list table ({s}.{t}) is not a key".format(n=self.NC_NAME, s=sname, t=tname))

            val_col = glt.getColumn(self.GC_NAME)
            if val_col is None:
                raise ValueError(
                    'table specified for group lists ({s}.{t}) lacks a "{n}" column'.format(s=sname, t=tname,
                                                                                            n=self.GC_NAME))
        if glt == {}:
            return None
        else:
            return glt

    def set_node_acl(self, node, spec):
        node.acls.clear()
        acl_name = spec.get("acl")
        if acl_name is not None:
            self.add_node_acl(node, acl_name)

    def expand_groups(self):
        for group_name in self.groups.keys():
            self.expand_group(group_name)

    def get_group(self, group_name):
        group = self.groups.get(group_name)
        if group is None:
            group = [group_name]
        return group

    def validate_group(self, group):
        if group == '*':
            return
        elif group.startswith(self.GLOBUS_PREFIX):
            self.validate_globus_group(group)
        elif group.startswith(self.ROBOT_PREFIX_FORMAT.format(server=self.server)):
            self.validate_webauthn_robot(group)
        else:
            warnings.warn("Can't determine format of group '{g}'".format(g=group))

    def validate_globus_group(self, group):
        guid = group[len(self.GLOBUS_PREFIX):]
        try:
            UUID(guid)
        except ValueError:
            raise ValueError("Group '{g}' appears to be a malformed Globus group".format(g=group))
        if self.verbose:
            print("group '{g}' appears to be a syntactically-correct Globus group".format(g=group))

    def validate_webauthn_robot(self, group):
        robot_name = group[len(self.ROBOT_PREFIX_FORMAT.format(server=self.server)):]
        if not robot_name:
            raise ValueError("Group '{g}' appears to be a malformed webauthn robot identity".format(g=group))
        if self.verbose:
            print("group '{g}' appears to be a syntactically-correct webauthn robot identity".format(g=group))

    def expand_group(self, group_name):
        groups = []
        for child_name in self.groups.get(group_name):
            child = self.groups.get(child_name)
            if child is None:
                self.validate_group(child_name)
                groups.append(child_name)
            else:
                self.expand_group(child_name)
                groups = groups + self.groups[child_name]
        self.groups[group_name] = list(set(groups))

    def expand_acl_definitions(self):
        for acl_name in self.acl_definitions.keys():
            self.expand_acl_definition(acl_name)

    def expand_acl_definition(self, acl_name):
        spec = self.acl_definitions.get(acl_name)
        for op_type in spec.keys():
            groups = []
            raw_groups = spec[op_type]
            if isinstance(raw_groups, list):
                for group_name in spec[op_type]:
                    groups = groups + self.get_group(group_name)
            else:
                groups = self.get_group(raw_groups)
            spec[op_type] = groups

    def set_table_acls(self, table):
        spec = self.acl_specs["table_acls"].find_best_table_spec(table.schema.name, table.name)
        table.acls.clear()
        table.acl_bindings.clear()
        if spec is not None:
            self.set_node_acl(table, spec)
            self.set_node_acl_bindings(table, table, spec.get("acl_bindings"), spec.get("invalidate_bindings"))
        if self.verbose:
            print(
                "set table {s}.{t} acls to {a}, bindings to {b}".format(s=table.schema.name, t=table.name, a=str(table.acls),
                                                                        b=str(table.acl_bindings)))
        for column in table.column_definitions:
            self.set_column_acls(column, table)
        for fkey in table.foreign_keys:
            self.set_fkey_acls(fkey, table)

    def set_column_acls(self, column, table):
        spec = self.acl_specs["column_acls"].find_best_column_spec(column.table.schema.name, column.table.name, column.name)
        column.acls.clear()
        column.acl_bindings.clear()
        if spec is not None:
            self.set_node_acl(column, spec)
            self.set_node_acl_bindings(column, table, spec.get("acl_bindings"), spec.get("invalidate_bindings"))
        if self.verbose:
            print("set column {s}.{t}.{c} acls to {a}, bindings to {b}".format(s=column.table.schema.name, t=column.table.name,
                                                                               c=column.name, a=str(column.acls),
                                                                               b=str(column.acl_bindings)))

    def set_fkey_acls(self, fkey, table):
        spec = self.acl_specs["foreign_key_acls"].find_best_foreign_key_spec(fkey.table.schema.name, fkey.table.name, fkey.names)
        fkey.acls.clear()
        fkey.acl_bindings.clear()
        if spec is not None:
            self.set_node_acl(fkey, spec)
            self.set_node_acl_bindings(fkey, table, spec.get("acl_bindings"), spec.get("invalidate_bindings"))
        if self.verbose:
            print("set fkey {f} acls to {a}, bindings to {b}".format(f=str(fkey.names), a=str(fkey.acls),
                                                                     b=str(fkey.acl_bindings)))

    def set_catalog_acls(self, catalog):
        spec = self.acl_specs["catalog_acl"]
        if spec is not None:
            catalog.acls.clear()
            self.set_node_acl(catalog, spec)
        if self.verbose:
            print("set catalog acls to {a}".format(a=str(catalog.acls)))
        for schema in self.toplevel_config.schemas.values():
            self.set_schema_acls(schema)

    def set_schema_acls(self, schema):
        for pattern in self.ignored_schema_patterns:
            if pattern.match(schema.name) is not None:
                print("ignoring schema {s}".format(s=schema.name))
                return
        spec = self.acl_specs["schema_acls"].find_best_schema_spec(schema.name)
        schema.acls.clear()
        if spec is not None:
            self.set_node_acl(schema, spec)
        if self.verbose:
            print("set schema {s} acls to {a}".format(s=schema.name, a=str(schema.acls)))

        for table in schema.tables.values():
            self.set_table_acls(table)

    def set_acls(self):
        if isinstance(self.toplevel_config, ermrest_model.Model):
            self.set_catalog_acls(self.toplevel_config)
        elif isinstance(self.toplevel_config, ermrest_model.Schema):
            self.set_schema_acls(self.toplevel_config)
        elif isinstance(self.toplevel_config, ermrest_model.Table):
            self.set_table_acls(self.toplevel_config)
        else:
            raise ValueError("toplevel config is a {t}".format(t=str(type(self.toplevel_config))))

    def apply_acls(self):
        self.toplevel_config.apply(self.saved_toplevel_config)

    def dumps(self):
        """Dump a serialized (string) representation of the config.
        """
        return json.dumps(self.toplevel_config.prejson(), indent=2)


class Table(AttrDict):
    ERMREST_DEFAULT_COLS = ["RID", "RCB", "RMB", "RCT", "RMT"]

    def __init__(self, d):
        if d is None:
            return
        self.base_entity_url = "/entity/{s}:{t}".format(s=d['schema_name'], t=d['table_name'])
        AttrDict.__init__(self, d)

    def getColumn(self, name):
        if self.get('column_definitions') is None:
            return None
        for c in self['column_definitions']:
            if c.get('name') == name:
                return c
        return None

    def getBaseEntityURL(self):
        return self.base_entity_url

    def upsertRows(self, catalog, rows):
        try:
            self.insertRows(catalog, rows)
        except HTTPError as err:
            if err.response.status_code == 409:
                for row in rows:
                    self.upsertRow(catalog, row)

    def find_keys(self):
        keys = self.get('keys')
        if keys is None:
            return keys
        for k in keys:
            for u in k.get('unique_columns'):
                c = self.getColumn(u)
                if c.get('nullok'):
                    keys.remove(k)
                    break
        return keys

    def row_has_key(self, row, key):
        for u in key.get('unique_columns'):
            if row.get(u) is None:
                return False
        return True

    def getRowFilter(self, row):
        filters = []
        key = None
        for k in self.find_keys():
            if self.row_has_key(row, k):
                key = k
                break

        if key is None:
            raise ValueError("can't find appropriate key")
        for k in key.get('unique_columns'):
            filters.append("{k}={v}".format(k=urlquote(k), v=urlquote(row[k])))
        return filters

    def getRow(self, catalog, row, filters):
        url = "{u}/{f}".format(u=self.getBaseEntityURL(), f="&".join(filters))
        vals = catalog.get(url, headers={'Content-Type': 'application/json'}).json()
        if vals is None or len(vals) == 0:
            return None
        return vals[0]

    def getDefaultCols(self, add_ermrest_defaults=True):
        default_cols = []
        for col in self.column_definitions:
            if col.get("default") is not None:
                default_cols.append(col.get("name"))
        if add_ermrest_defaults:
            default_cols = list(set(default_cols + self.ERMREST_DEFAULT_COLS))
        return default_cols

    def upsertRow(self, catalog, row):
        try:
            return self.insertRows(catalog, [row])
        except HTTPError as err:
            if err.response.status_code == 409:
                return self.updateRow(catalog, row)

    def updateRow(self, catalog, row):
        filters = self.getRowFilter(row)
        old_row = self.getRow(catalog, row, filters)
        for c in self.getDefaultCols():
            if row.get(c) is None and old_row.get(c) is not None:
                row[c] = old_row[c]
        return catalog.put(self.getBaseEntityURL(), json=[row], headers={'Content-Type': 'application/json'})

    def insertRows(self, catalog, rows):
        default_cols = self.getDefaultCols(False)
        if default_cols is not None and len(default_cols) != 0:
            url = "{u}?defaults={d}".format(u=self.getBaseEntityURL(), d=",".join(default_cols))
        else:
            url = self.getBaseEntityURL()
        return catalog.post(url, json=rows, headers={'Content-Type': 'application/json'})

    def __str__(self):
        return dict.__str__(self)


class AclCLI(ConfigBaseCLI):
    def __init__(self):
        ConfigBaseCLI.__init__(self, "ACL configuration tool", None, version=VERSION)
        group = self.parser.add_mutually_exclusive_group()
        group.add_argument('-g', '--groups-only', help="create group table only", action="store_true")
        group.add_argument('-o', '--omit-groups', help="do not create group table", action="store_true")


def main():
    cli = AclCLI()
    args = cli.parse_cli()
    table_name = cli.get_table_arg(args)
    schema_names = cli.get_schema_arg_list(args)
    credentials = get_credential(args.host, args.credential_file)
    save_groups = not (args.dryrun or args.omit_groups)
    for schema in schema_names:
        acl_config = AclConfig(args.host, args.catalog, args.config_file, credentials, schema_name=schema,
                               table_name=table_name, verbose=args.verbose or args.debug)

        try:
            if save_groups:
                acl_config.save_groups()
                save_groups = False
            if not args.groups_only:
                acl_config.set_acls()
                if not args.dryrun:
                    acl_config.apply_acls()
        except HTTPError as e:
            print(format_exception(e))
            raise

        if args.dryrun:
            print(acl_config.dumps())


if __name__ == '__main__':
    sys.exit(main())



import logging
from deriva.transfer.download import DerivaDownloadConfigurationError, DerivaDownloadError
from deriva.transfer.download.processors.base_processor import *

logger = logging.getLogger(__name__)

INPUT_PATHS_KEY = 'input_paths'
REMOTE_PATH_KEY = 'remote_path'
OUTPUT_KEY = 'output'
OUTPUT_URLENCODED_KEY = 'output_urlencoded'
OUTPUT_URL_KEY = 'output_url'
OUTPUT_URL_URLENCODED_KEY = 'output_url_urlencoded'


class UrlRewritePostProcessor(BaseProcessor):
    """Post processor that rewrites the output remote paths.
    
    This can be used for scenarios where the client should be directed to a secondary service that consumes the export
    service's output file.

    Service params:
      `input_paths`: a list of input path names which will be post processed.
      `remote_path`: the remote path pattern formatted with the following env.

    Formatting environment:
      - copy of all `envars` in the current environment, plus
      - `output`: the current output key
      - `output_url`: the url to the current output formed by `deriva_service_url` + `/` + `output`
      - and `..._urlencoded` variants of the above using the deriva urlquote() function.
    """

    def __init__(self, envars=None, **kwargs):
        super(UrlRewritePostProcessor, self).__init__(envars, **kwargs)

        # get remote path pattern
        self.remote_path_pattern = self.parameters.get(REMOTE_PATH_KEY)
        if not self.remote_path_pattern:
            raise DerivaDownloadConfigurationError(
                    "%s is missing required parameter '%s' from %s" %
                    (self.__class__.__name__, REMOTE_PATH_KEY, PROCESSOR_PARAMS_KEY))
        logger.debug("remote path pattern: {}".format(self.remote_path_pattern))

        # get input paths
        self.input_paths = self.parameters.get(INPUT_PATHS_KEY)
        if not self.input_paths:
            raise DerivaDownloadConfigurationError(
                    "%s is missing required parameter '%s' from %s" %
                    (self.__class__.__name__, INPUT_PATHS_KEY, PROCESSOR_PARAMS_KEY))

    def process(self):
        env = self.envars.copy()
        for input_path in self.input_paths:
            k = input_path
            v = self.outputs.get(k)
            if not v:
                raise DerivaDownloadError(
                    "%s is missing required input path '%s'" % (self.__class__.__name__, k))
            env[OUTPUT_KEY] = k
            env[OUTPUT_URLENCODED_KEY] = urlquote(k)
            env[OUTPUT_URL_KEY] = self.envars.get(SERVICE_URL_KEY, '') + '/' + k
            env[OUTPUT_URL_URLENCODED_KEY] = urlquote(env[OUTPUT_URL_KEY])
            v[REMOTE_PATHS_KEY] = [self.remote_path_pattern.format(**env)]

        return self.outputs
